"""
Shim for global config access.
"""

from gi.repository import Gio
from enum import IntEnum

from . import APP_ID

config = Gio.Settings.new(APP_ID)


class DLCoverSize(IntEnum):
    NO_DOWNLOAD = 0
    _250PX = 250
    _500PX = 500
    _1200PX = 1200
    MAX_SIZE = 2000

    @classmethod
    def index_to_item(cls, i):
        if i == 0:
            return cls.NO_DOWNLOAD
        elif i == 1:
            return cls._250PX
        elif i == 2:
            return cls._500PX
        elif i == 3:
            return cls._1200PX
        elif i == 4:
            return cls.MAX_SIZE
        raise ValueError

    @classmethod
    def item_to_index(cls, o):
        o = int(o)

        if o == 0:
            return 0
        elif o == 250:
            return 1
        elif o == 500:
            return 2
        elif o == 1200:
            return 3
        elif o == 2000:
            return 4
        raise ValueError



# SPDX-License-Identifier: MIT
"""
Code for generic queued downloader.
"""

import asyncio
from aiohttp.client_exceptions import ClientConnectorError
import aiohttp
from aiohttp_retry import RetryClient, ExponentialRetry
from enum import Enum
from tempfile import NamedTemporaryFile
from typing import Union
import time
from ..logger import logger

try:
    from .. import VERSION
except ImportError:  # handle test suite import
    from tests.common import VERSION

    USER_AGENT = f"Ear Tag (Test suite)/{VERSION} (https://gitlab.gnome.org/World/eartag)"
else:
    USER_AGENT = f"Ear Tag/{VERSION} (https://gitlab.gnome.org/World/eartag)"


CHUNK_SIZE = 1024 * 1024  # 1MB


class EartagDownloaderMode(Enum):
    MODE_TEXT = 0
    MODE_JSON = 1
    MODE_FILE = 2


class EartagQueuedDownloader:
    """
    Generic queued downloader. Fetches data from a remote URL and can either
    store it in a cache (for JSON/text data) or save to a tempfile (for files).
    """

    def __init__(
        self,
        mode: EartagDownloaderMode,
        simultaneous_downloads: int = 3,
        throttle: bool = False,
    ):
        """Initializes a queued downloader."""
        self.mode = mode
        # Key - URL, value - data. Value of False is equivalent to an error.
        self.cache = {}
        self.semaphore = asyncio.Semaphore(simultaneous_downloads)
        self.throttle = throttle
        self._last_request = time.time()

    def __del__(self):
        try:
            for val in self.cache.values():
                del val
            del self.cache
        except AttributeError:
            pass

    async def download(self, url: str):
        """Perform a download for the given URL."""
        logger.debug(f"QueuedDL: Downloading {url}...")

        if url in self.cache:
            # If the URL is cached, just return the data
            logger.debug(f"QueuedDL: Found {url} in cache")
            return self.get_cached(url)

        retry_options = ExponentialRetry(
            attempts=3,
            exceptions={ClientConnectorError},
            factor=1,
            statuses={500, 502, 503, 504},
        )

        async with self.semaphore:
            if self.throttle:
                cur_time = time.time()
                if (cur_time - self._last_request) < 1000:
                    await asyncio.sleep((1000 - (cur_time - self._last_request)) / 1000)

            async with aiohttp.ClientSession() as session:
                async with RetryClient(
                    client_session=session, retry_options=retry_options
                ) as retry_session:
                    async with retry_session.get(
                        url, timeout=60, headers={"User-Agent": USER_AGENT}
                    ) as response:
                        if response.status != 200:
                            data = False
                        else:
                            if self.mode == EartagDownloaderMode.MODE_TEXT:
                                data = await response.text()

                            elif self.mode == EartagDownloaderMode.MODE_JSON:
                                try:
                                    data = await response.json()
                                except:  # noqa: E722
                                    data = False

                            elif self.mode == EartagDownloaderMode.MODE_FILE:
                                tempfile = NamedTemporaryFile()
                                async for chunk in response.content.iter_chunked(CHUNK_SIZE):
                                    tempfile.write(chunk)
                                tempfile.flush()
                                data = tempfile

                        self.cache[url] = data

        if data is False:
            logger.debug(f"QueuedDL: Download for {url} failed")
        else:
            logger.debug(f"QueuedDL: Download for {url} succeeded")
        return data

    def get_cached(self, url: str) -> Union[str, dict, NamedTemporaryFile, bool]:
        """
        Get the value of a cached URL.

        The output value type depends on the selected mode:

        * MODE_TEXT: str
        * MODE_JSON: dict
        * MODE_FILE: NamedTemporaryFile

        If the URL failed to download or was not found on the server,
        False is returned. If the URL is not cached, None is returned.
        """
        return self.cache.get(url, None)

    def clear_tempfiles(self):
        """For queued downloaders in file mode, close all tempfiles."""
        if not self.mode == EartagDownloaderMode.MODE_FILE:
            return

        for tmp in self.cache.values():
            if tmp:
                tmp.close()

        del self.cache



# SPDX-License-Identifier: MIT
# (c) 2023 knuxify and Ear Tag contributors

from gi.repository import Adw, GObject, Gtk

from ._async import event_loop
from .utils.limiters import EartagEntryLimiters
from .utils.widgets import EartagEditableLabel
from .utils.misc import all_equal, safe_int, safe_float
from .backends.file import EartagFile

from collections.abc import Iterable


class EartagTagEntryManager(GObject.Object):
    """
    Manager grouping together multiple TagEntry-compatible entries.
    Handles binding to a file and synchronizing the values between
    the entries and the bound files.
    """

    def __init__(self):
        """
        Initialize the necessary internal variables for the manager.

        Should be called in the __init__ function of derived classes.
        """
        super().__init__()

        #: Mapping of property to entry.
        self.entries: dict[str, Gtk.Widget] = {}

        #: List of connections made to the entry.
        self.entry_connections: dict[str, list[int]] = {}

        #: Current inconsistency state for the entry.
        self.entry_inconsistency: dict[str, bool] = {}

        #: Set of files bound to the manager.
        self.files: set = set()

        #: List of connections made to files. The key is the file ID.
        self.file_connections: dict[str, list[int]] = {}

        #: If True, the file tag modification signal is ignored.
        self._ignore_file_change: bool = False

        #: If True, the entry text change signal is ignored.
        self._ignore_entry_change: bool = False

    @property
    def managed_properties(self) -> set:
        """Get the set of properties this manager has entries for."""
        return set(self.entries.keys())

    def destroy(self):
        """
        Remove all entries and unbind from all files.

        Should be called from the destroy signal of the derived class.
        """
        event_loop.create_task(self.unbind_from_files(self.files.copy()))

        for prop in self.managed_properties:
            self.remove_entry(prop)

    # File management

    async def bind_to_files(self, files: Iterable[EartagFile]):
        for file in files:
            # TODO: Is this faster than an "if file in self.files" check?
            n_files = len(self.files)
            self.files.add(file)
            if len(self.files) == n_files:
                return
            n_files += 1

            self.file_connections[file.id] = [file.connect("modified", self.handle_file_modified)]

            # Update the entries' "inconsistent" (multiple values) state
            if n_files > 1:
                for prop in [k for k, v in self.entry_inconsistency.items() if v is False]:
                    if prop not in self.entries:
                        continue
                    entry = self.entries[prop]
                    if prop in file.int_properties:
                        val = safe_int(entry.props.text)
                    elif prop in file.float_properties:
                        val = safe_float(entry.props.text)
                    else:
                        val = entry.props.text

                    self.entry_inconsistency[prop] = val != file.get_property(prop)
            else:
                for prop in self.managed_properties:
                    self.entry_inconsistency[prop] = False

        for prop in self.managed_properties:
            self.refresh_entry_text(prop)

    async def unbind_from_files(self, files: Iterable[EartagFile]):
        """Unbind from the files provided by the iterable."""
        for file in files:
            if file.id not in self.file_connections:
                return
            for conn in self.file_connections[file.id]:
                file.disconnect(conn)
            self.files.remove(file)

        for prop in self.managed_properties:
            self.recalculate_inconsistent_state(prop)
            self.refresh_entry_text(prop)

    @GObject.Property(type=bool, default=False)
    def is_busy(self):
        """Whether the manager is currently binding/unbinding files or not."""
        return self.bind_task.is_running or self.unbind_task.is_running

    def handle_file_modified(self, file: EartagFile, tag: str):
        """Handle a file's value being modified."""
        if self._ignore_file_change:
            return

        if tag in self.managed_properties:
            self.refresh_entry_text(tag)

    def recalculate_inconsistent_state(self, prop: str):
        """
        Recalculate the entry inconsistency state for the entry with
        the given property.
        """
        self.entry_inconsistency[prop] = not all_equal(
            file.get_property(prop) for file in self.files
        )

    # Entry management

    def add_entry(self, prop: str, entry: Gtk.Widget):
        """Add an entry for the given property."""
        if prop in self.entries:
            return

        if entry.bound_property != prop:
            entry.bound_property = prop

        self.entries[prop] = entry

        self.entry_connections[prop] = [
            entry.connect("notify::bound-property", self.handle_entry_prop_changed, prop),
            entry.connect("changed", self.handle_entry_changed, prop),
        ]
        self.entry_inconsistency[prop] = False

        self.refresh_entry_text(prop)

    def remove_entry(self, prop: str):
        """Remove the entry for the given property."""
        if prop not in self.entries:
            return

        for conn in self.entry_connections[prop]:
            self.entries[prop].disconnect(conn)
        del self.entry_connections[prop]

        del self.entry_inconsistency[prop]
        del self.entries[prop]

    def get_entry_by_property(self, prop: str) -> Gtk.Widget | None:
        """Return the entry for this property."""
        return self.entries.get(prop, None)

    def has_entry_for_property(self, prop: str) -> bool:
        """Check whether an entry for the given property is present."""
        return prop in self.entries

    def handle_entry_prop_changed(self, entry: Gtk.Widget, old_prop: str):
        """Handle the bound_property value of an entry being changed."""
        new_prop = entry.bound_property

        self.remove_entry(old_prop)
        self.add_entry(entry, new_prop)

    def handle_entry_changed(self, entry: Gtk.Widget, prop: str):
        """Handle the entry contents being changed."""
        if self._ignore_entry_change:
            return

        self._ignore_file_change = True

        entry.tagentry_placeholder = ""

        if prop in EartagFile.int_properties:
            value = safe_int(entry.props.text)
        elif prop in EartagFile.float_properties:
            value = safe_float(entry.props.text)
        else:
            value = entry.props.text

        for file in self.files:
            file.set_property(prop, value)

        self.entry_inconsistency[prop] = False

        self._ignore_file_change = False

    def refresh_entry_text(self, prop: str):
        """Refresh the data for the entry with the given property."""
        self._ignore_entry_change = True

        entry = self.entries[prop]
        if self.entry_inconsistency[prop]:
            entry.props.text = ""
            entry.tagentry_placeholder = _("(multiple values)")
        else:
            try:
                entry.props.text = str(next(iter(self.files)).get_property(prop) or "")
                entry.tagentry_placeholder = ""
            except StopIteration:
                entry.props.text = ""
                entry.tagentry_placeholder = ""

        self._ignore_entry_change = False


class EartagTagEntryBase(GObject.Object):
    """
    Base class for tag entries: entries which are bound to files and carry
    the value for a specific tag.

    Entries are expected to have the following:
     - "text" property
     - "changed" signal
     - "tagentry-placeholder" property
    The first two are implemented by GtkEditable; the last one has to be
    implemented in the entry class manually.
    """

    # The following line needs to be copied manually to the derived class
    # (until https://gitlab.gnome.org/GNOME/pygobject/-/issues/577 is done)
    # bound_property = GObject.Property(type=str, default=None)


class EartagTagEntry(Gtk.Entry, EartagTagEntryBase, EartagEntryLimiters):
    """Simple GtkEntry implementing EartagTagEntryBase."""

    __gtype_name__ = "EartagTagEntry"

    def __init__(self):
        super().__init__()
        self.setup_limiters()

    @GObject.Property(type=str)
    def tagentry_placeholder(self):
        return self.props.placeholder_text

    @tagentry_placeholder.setter
    def tagentry_placeholder(self, value):
        # Bit of a hack, but: override the placeholder value for track
        # number and total track number entries to prevent them from
        # expanding the app window when set
        if value and self.bound_property in ("tracknumber", "totaltracknumber"):
            self.props.placeholder_text = "..."
        else:
            self.props.placeholder_text = value

    # https://gitlab.gnome.org/GNOME/pygobject/-/issues/577
    bound_property = GObject.Property(type=str, default=None)

    @GObject.Property(type=bool, default=False)
    def is_numeric(self):
        return self._is_numeric_prop

    @is_numeric.setter
    def is_numeric(self, value):
        self._is_numeric_prop = value

    @GObject.Property(type=bool, default=False)
    def is_float(self):
        return self._is_float_prop

    @is_float.setter
    def is_float(self, value):
        self._is_float_prop = value

    @GObject.Property(type=bool, default=False)
    def is_date(self):
        return self._is_date_prop

    @is_date.setter
    def is_date(self, value):
        self._is_date_prop = value


class EartagTagEntryRow(Adw.EntryRow, EartagTagEntryBase, EartagEntryLimiters):
    """Simple AdwEntryRow implementing EartagTagEntryBase."""

    __gtype_name__ = "EartagTagEntryRow"

    def __init__(self):
        super().__init__()
        self._placeholder_text = ""
        self._title = self.props.title
        self.setup_limiters()

    # AdwEntryRows do not have placeholder text; the entry's title doubles
    # as the placeholder. So, to simulate a placeholder without ruining
    # the entry title, we append the placeholder text to the title.
    @GObject.Property(type=str)
    def tagentry_placeholder(self):
        return self._placeholder_text

    @tagentry_placeholder.setter
    def tagentry_placeholder(self, value):
        if not self._title and self.props.title:
            self._title = self.props.title
        self._placeholder_text = value
        if value:
            self.set_title(self._title + " " + value)
        else:
            self.set_title(self._title)

    # https://gitlab.gnome.org/GNOME/pygobject/-/issues/577
    bound_property = GObject.Property(type=str, default=None)

    @GObject.Property(type=bool, default=False)
    def is_numeric(self):
        return self._is_numeric_prop

    @is_numeric.setter
    def is_numeric(self, value):
        self._is_numeric_prop = value

    @GObject.Property(type=bool, default=False)
    def is_float(self):
        return self._is_float_prop

    @is_float.setter
    def is_float(self, value):
        self._is_float_prop = value

    @GObject.Property(type=bool, default=False)
    def is_date(self):
        return self._is_date_prop

    @is_date.setter
    def is_date(self, value):
        self._is_date_prop = value


class EartagTagEditableLabel(EartagEditableLabel, EartagTagEntryBase):
    """Simple EartagEditableLabel implementing EartagTagEntryBase."""

    __gtype_name__ = "EartagTagEditableLabel"

    def __init__(self):
        super().__init__()
        self._tagentry_placeholder = ""

    @GObject.Property(type=str, default=None)
    def title(self):
        return self._title

    @title.setter
    def title(self, value):
        self._title = value
        Gtk.Accessible.update_property(self, (Gtk.AccessibleProperty.LABEL,), (value,))
        Gtk.Accessible.update_property(self.entry, (Gtk.AccessibleProperty.LABEL,), (value,))

    # Custom handler to avoid overwriting the original placeholder
    @GObject.Property(type=str, default=None)
    def tagentry_placeholder(self):
        return self._tagentry_placeholder

    @tagentry_placeholder.setter
    def tagentry_placeholder(self, value):
        self._tagentry_placeholder = value
        if value:
            self.placeholder_text = self._title + " " + value
        else:
            self.placeholder_text = self._title

    bound_property = GObject.Property(type=str, default=None)



# SPDX-License-Identifier: MIT
# (c) 2023 knuxify and Ear Tag contributors

from gi.repository import GObject
import asyncio
import base64
from PIL import Image
import io

import mutagen
from mutagen.flac import FLAC, Picture, error as FLACError
from mutagen.id3 import PictureType

from .file import CoverType
from .file_mutagen_common import EartagFileMutagenCommon
from ..utils.misc import safe_int
from ..utils.validation import get_mimetype_buffer


class EartagFileMutagenVorbis(EartagFileMutagenCommon):
    """EartagFile handler that uses mutagen for Vorbis Comment support."""

    __gtype_name__ = "EartagFileMutagenVorbis"
    _supports_album_covers = True
    _cover_mimetypes = ["image/jpeg", "image/png"]
    _supports_full_dates = True

    # There's an official standard and semi-official considerations for tags,
    # plus some more documents linked from https://wiki.xiph.org/VorbisComment;
    # this only covers tags mentioned there.
    supported_extra_tags = (
        "composer", "copyright", "encodedby", "mood", "discnumber", "publisher",
        "isrc",

        "albumartistsort", "albumsort", "composersort", "artistsort", "titlesort",

        "musicbrainz_artistid", "musicbrainz_albumid",
        "musicbrainz_albumartistid", "musicbrainz_trackid",
        "musicbrainz_recordingid", "musicbrainz_releasegroupid"
    )  # fmt: skip

    _replaces = {
        "totaltracknumber": "tracktotal",
        "releasedate": "date",
        # There"s also ENCODED-BY, but confusingly it represents... the person doing the encoding?
        "encodedby": "encoder",
        # Matching Picard behavior:
        "musicbrainz_trackid": "musicbrainz_releasetrackid",
        "musicbrainz_recordingid": "musicbrainz_trackid"
    }  # fmt: skip

    async def load_from_file(self, path):
        from mutagen.oggflac import OggFLAC
        from mutagen.oggspeex import OggSpeex
        from mutagen.oggtheora import OggTheora
        from mutagen.oggvorbis import OggVorbis
        from mutagen.oggopus import OggOpus

        self.mg_file = await asyncio.to_thread(
            mutagen.File,
            path,
            options=[OggTheora, OggSpeex, OggVorbis, OggOpus, OggFLAC, FLAC],
        )
        if self.mg_file is None:
            raise ValueError("Failed to identify OGG subtype")

        await self.load_cover()
        self.setup_present_extra_tags()
        self.setup_original_values()

    def get_tag(self, tag_name):
        """Tries the lowercase, then uppercase representation of the tag."""
        if tag_name.lower() in self._replaces:
            tag_name = self._replaces[tag_name.lower()]
        try:
            if tag_name in self.int_properties:
                return self.mg_file.tags[tag_name.lower()][0]
            else:
                return self.mg_file.tags[tag_name.lower()][0] or ""
        except KeyError:
            try:
                if tag_name in self.int_properties:
                    return self.mg_file.tags[tag_name.upper()][0]
                else:
                    return self.mg_file.tags[tag_name.upper()][0] or ""
            except KeyError:
                if tag_name in self.int_properties:
                    return None
                else:
                    return ""

    def set_tag(self, tag_name, value):
        """Sets the tag with the given name to the given value."""
        if tag_name.lower() in self._replaces:
            tag_name = self._replaces[tag_name.lower()]
        if tag_name.upper() in self.mg_file.tags:
            self.mg_file.tags[tag_name.upper()] = str(value)
        else:
            self.mg_file.tags[tag_name] = str(value)

    def has_tag(self, tag_name):
        """
        Returns True or False based on whether the tag with the given name is
        present in the file.
        """
        if tag_name == "totaltracknumber":
            return bool(self.totaltracknumber)
        elif tag_name == "encodedby":
            if "encoder" in self.mg_file.tags:
                return bool(self.mg_file.tags["encoder"][0])
            elif "ENCODER" in self.mg_file.tags:
                return bool(self.mg_file.tags["ENCODER"][0])
            else:
                return False
        if tag_name.lower() in self._replaces:
            tag_name = self._replaces[tag_name.lower()]
        if tag_name in self.mg_file.tags or tag_name.upper() in self.mg_file.tags:
            return True
        return False

    def delete_tag(self, tag_name):
        """Deletes the tag with the given name from the file."""
        _original_tag_name = tag_name
        if tag_name.lower() == "releasedate":
            self._releasedate_cached = ""
        if tag_name.lower() in self._replaces:
            tag_name = self._replaces[tag_name.lower()]
        if tag_name in self.mg_file.tags:
            del self.mg_file.tags[tag_name]
        self.mark_as_modified(_original_tag_name, notify_prop=True)

    def delete_cover(self, cover_type: CoverType, clear_only=False):
        if cover_type == CoverType.FRONT:
            pictypes = (PictureType.OTHER, PictureType.COVER_FRONT)
        elif cover_type == CoverType.BACK:
            pictypes = (PictureType.COVER_BACK,)
        else:
            raise ValueError

        if isinstance(self.mg_file, FLAC):
            pic_list = list(self.mg_file.pictures)
            for _pic in pic_list.copy():
                if _pic.type in pictypes:
                    pic_list.remove(_pic)

            # There's no way to remove a picture, so we have to clear and re-add
            # all the existing ones:
            self.mg_file.clear_pictures()
            for _pic in pic_list:
                self.mg_file.add_picture(_pic)

        pic_list = list(self.mg_file.get("metadata_block_picture", []))
        for b64_data in pic_list.copy():
            try:
                data = base64.b64decode(b64_data)
            except (TypeError, ValueError):
                continue

            try:
                cover_picture = Picture(data)
            except FLACError:
                continue

            if cover_picture.type in pictypes:
                pic_list.remove(b64_data)

        self.mg_file["metadata_block_picture"] = pic_list

        if not clear_only:
            self._cleanup_cover(cover_type)

    def on_remove(self, *args):
        super().on_remove()

    async def set_cover_from_data(self, cover_type: CoverType, data: str, mime: str | None = None):
        if cover_type == CoverType.FRONT:
            pictype = PictureType.COVER_FRONT
        elif cover_type == CoverType.BACK:
            pictype = PictureType.COVER_BACK
        else:
            raise ValueError

        if not mime:
            mime = get_mimetype_buffer(data)

        # Set cover in UI and check if it's valid
        ret = await self._set_cover_from_data(cover_type, data)
        if ret is False:
            return

        # shamelessly stolen from
        # https://stackoverflow.com/questions/1996577/how-can-i-get-the-depth-of-a-jpg-file
        mode_to_bpp = {
            "1": 1,
            "L": 8,
            "P": 8,
            "RGB": 24,
            "RGBA": 32,
            "CMYK": 32,
            "YCbCr": 24,
            "LAB": 24,
            "HSV": 24,
            "I": 32,
            "F": 32,
        }

        picture = Picture()
        picture.data = data
        picture.type = pictype
        picture.mime = mime

        def _get_metadata(data, picture):
            with Image.open(io.BytesIO(data)) as img:
                picture.width = img.width
                picture.height = img.height
                picture.depth = mode_to_bpp[img.mode]

        await asyncio.to_thread(_get_metadata, data, picture)

        if isinstance(self.mg_file, FLAC):
            self.mg_file.add_picture(picture)

        picture_data = picture.write()
        encoded_data = base64.b64encode(picture_data)
        vcomment_value = encoded_data.decode("ascii")
        if "metadata_block_picture" in self.mg_file:
            self.mg_file["metadata_block_picture"] = [vcomment_value] + self.mg_file[
                "metadata_block_picture"
            ]
        else:
            self.mg_file["metadata_block_picture"] = [vcomment_value]

    async def _load_cover(self, cover_type: CoverType):
        """Loads cover data from file."""

        # See https://mutagen.readthedocs.io/en/latest/user/vcomment.html.
        # There are three ways to get the cover image:

        # 1. Using `mutagen.flac.FLAC.pictures`
        if isinstance(self.mg_file, FLAC) and self.mg_file.pictures:
            picture = None
            picture_cover = None
            picture_other = None
            picture_back = None

            if cover_type == CoverType.FRONT:
                for _picture in self.mg_file.pictures:
                    if _picture.type == PictureType.COVER_FRONT:
                        picture_cover = _picture

                    elif _picture.type == PictureType.OTHER:
                        picture_other = _picture

                if picture_cover:
                    picture = picture_cover
                elif picture_other:
                    picture = picture_other

                if picture:
                    await self._set_cover_from_data(cover_type, picture.data, modified=False)

            elif cover_type == CoverType.BACK:
                for _picture in self.mg_file.pictures:
                    if _picture.type == PictureType.COVER_BACK:
                        picture_back = _picture
                        break

                if picture_back:
                    await self._set_cover_from_data(cover_type, picture_back.data, modified=False)

        # 2. Using metadata_block_picture
        elif self.mg_file.get("metadata_block_picture", []):
            cover_front = None
            covers_other = []
            cover_back = None

            for b64_data in self.mg_file.get("metadata_block_picture", []):
                try:
                    data = base64.b64decode(b64_data)
                except (TypeError, ValueError):
                    continue

                try:
                    cover_picture = Picture(data)
                except FLACError:
                    continue

                if cover_type == CoverType.FRONT:
                    if cover_picture.type == PictureType.COVER_FRONT:
                        cover_front = cover_picture
                        break
                    elif cover_picture.type == PictureType.OTHER:
                        covers_other.append(cover_picture)

                elif cover_type == CoverType.BACK:
                    if cover_picture.type == PictureType.COVER_BACK:
                        cover_back = cover_picture
                        break

            if not cover_front and covers_other:
                cover_front = covers_other[0]

            if cover_front:
                await self._set_cover_from_data(CoverType.FRONT, cover_front.data, modified=False)

            if cover_back:
                await self._set_cover_from_data(CoverType.BACK, cover_back.data, modified=False)

        # 3. Using the coverart field (and optionally covermime)
        elif cover_type == CoverType.FRONT:
            covers = self.mg_file.get("coverart", [])

            n = 0
            for cover in covers:
                try:
                    data = base64.b64decode(cover.encode("ascii"))
                except (TypeError, ValueError):
                    continue

                if not data:
                    continue

                await self._set_cover_from_data(cover_type, data, modified=False)
                n += 1

    async def load_cover(self):
        for cover_type in (CoverType.FRONT, CoverType.BACK):
            await self._load_cover(cover_type)

    @GObject.Property(type=int)
    def tracknumber(self):
        tracknum_raw = self.get_tag("tracknumber")
        if not tracknum_raw:
            return None
        if "/" in tracknum_raw:
            return safe_int(tracknum_raw.split("/")[0])

        return safe_int(tracknum_raw)

    @tracknumber.setter
    def tracknumber(self, value):
        if value:
            self.set_tag("tracknumber", str(value))
        else:
            self.delete_tag("tracknumber")
        self.mark_as_modified("tracknumber")

    @GObject.Property(type=int)
    def totaltracknumber(self):
        totalnum_raw = self.get_tag("totaltracknumber")
        if totalnum_raw:
            return safe_int(totalnum_raw)

        # Fall back to parsing track number with "/" (older Ear Tag versions)
        tracknum_raw = self.get_tag("tracknumber")
        if not tracknum_raw:
            return None
        if "/" in tracknum_raw:
            try:
                return safe_int(tracknum_raw.split("/")[1])
            except (IndexError, ValueError):
                return None

        return None

    @totaltracknumber.setter
    def totaltracknumber(self, value):
        # Fix up track numbers with "/"
        if "/" in (self.get_tag("tracknumber") or ""):
            if self.tracknumber:
                self.set_tag("tracknumber", str(self.tracknumber))
            else:
                self.delete_tag("tracknumber")
            self.mark_as_modified("tracknumber")

        if value:
            self.set_tag("totaltracknumber", str(value))
        else:
            self.delete_tag("totaltracknumber")

        self.mark_as_modified("totaltracknumber")

    @GObject.Property(type=int)
    def discnumber(self):
        discnum_raw = self.get_tag("discnumber")
        if not discnum_raw:
            return None
        if "/" in discnum_raw:
            return safe_int(discnum_raw.split("/")[0])
        return safe_int(discnum_raw)

    @discnumber.setter
    def discnumber(self, value):
        if value:
            self.set_tag("discnumber", str(value))
        elif self.has_tag("discnumber"):
            self.delete_tag("discnumber")
        self.mark_as_modified("discnumber")



# SPDX-License-Identifier: MIT
# (c) 2023 knuxify and Ear Tag contributors

from .config import config, DLCoverSize
from .utils.asynctask import EartagAsyncTask
from .utils.validation import is_valid_music_file, VALID_AUDIO_MIMES
from .utils.tagselector import EartagTagSelectorButton  # noqa: F401
from .dialogs import (
    EartagErrorType,
    EartagErrorDialog,
    EartagCloseWarningDialog,
    EartagDiscardWarningDialog,
    EartagTagDeleteWarningDialog,
)
from .fileview import EartagFileView  # noqa: F401
from .filemanager import EartagFileManager
from .filelist import EartagFileList, EartagFileListItem  # noqa: F401
from . import APP_GRESOURCE_PATH, DEVEL

import asyncio
from gi.repository import Adw, Gdk, GLib, Gtk, Gio, GObject
import os
from enum import Flag


class EartagFileDialogType(Flag):
    """Types of file opening dialogs."""

    # Dialog type - file or folder
    TYPE_FILE = 1
    TYPE_FOLDER = 2

    # Load mode - overwrite (clear all and load) or insert (append to current)
    MODE_OVERWRITE = 4
    MODE_INSERT = 8


@Gtk.Template(resource_path=f"{APP_GRESOURCE_PATH}/ui/nofile.ui")
class EartagNoFile(Adw.Bin):
    __gtype_name__ = "EartagNoFile"

    nofile_status = Gtk.Template.Child()
    open_file = Gtk.Template.Child()
    open_recursive = Gtk.Template.Child()

    def __init__(self):
        super().__init__()
        if DEVEL:
            self.nofile_status.set_icon_name("app.drey.EarTag.Devel")
        config.bind(
            "open-folders-recursively",
            self.open_recursive,
            "active",
            flags=Gio.SettingsBindFlags.DEFAULT,
        )

    def grab_button_focus(self, *args):
        self.open_file.grab_focus()

    @Gtk.Template.Callback()
    def on_add_file(self, *args):
        window = self.get_native()
        window.show_file_chooser(
            EartagFileDialogType.TYPE_FILE | EartagFileDialogType.MODE_OVERWRITE
        )

    @Gtk.Template.Callback()
    def on_add_folder(self, *args):
        window = self.get_native()
        window.show_file_chooser(
            EartagFileDialogType.TYPE_FOLDER | EartagFileDialogType.MODE_OVERWRITE
        )


@Gtk.Template(resource_path=f"{APP_GRESOURCE_PATH}/ui/window.ui")
class EartagWindow(Adw.ApplicationWindow):
    __gtype_name__ = "EartagWindow"

    save_button = Gtk.Template.Child()
    window_title = Gtk.Template.Child()
    container_stack = Gtk.Template.Child()
    split_view = Gtk.Template.Child()

    sidebar_view = Gtk.Template.Child()
    sidebar_headerbar = Gtk.Template.Child()
    sidebar_search_button = Gtk.Template.Child()
    select_multiple_button = Gtk.Template.Child()

    no_file = Gtk.Template.Child()
    file_view = Gtk.Template.Child()

    no_file_widget = Gtk.Template.Child()

    toast_overlay = Gtk.Template.Child()
    overlay = Gtk.Template.Child()
    drop_highlight_revealer = Gtk.Template.Child()

    primary_menu_button = Gtk.Template.Child()
    empty_primary_menu_button = Gtk.Template.Child()

    force_close = False
    file_chooser_mode = None

    # Sidebar
    sidebar_list_stack = Gtk.Template.Child()
    sidebar_list_scroll = Gtk.Template.Child()
    sidebar_file_list = Gtk.Template.Child()
    sidebar_no_files = Gtk.Template.Child()

    search_bar = Gtk.Template.Child()
    search_entry = Gtk.Template.Child()
    no_results = Gtk.Template.Child()

    sidebar_action_bar = Gtk.Template.Child()
    select_all_button = Gtk.Template.Child()
    remove_selected_button = Gtk.Template.Child()
    selected_message_label = Gtk.Template.Child()

    loading_progressbar = Gtk.Template.Child()
    loading_progressbar_revealer = Gtk.Template.Child()

    def __init__(self, application, paths=None, devel=False):
        super().__init__(application=application, title=_("Ear Tag"))
        self._undo_all_data = {}
        self.error_dialog = None
        self._identify_imported = False
        self._redo_all_called = False

        if devel:
            self.add_css_class("devel")

        self.set_icon_name(application.get_application_id())

        self.file_chooser = Gtk.FileDialog(modal=True)

        self.audio_file_filter = Gtk.FileFilter()
        for mime in VALID_AUDIO_MIMES:
            self.audio_file_filter.add_mime_type(mime)
        self.audio_file_filter.set_name(_("All supported audio files"))

        self.file_view.connect("notify::bind-in-progress", self.update_busy)

        # File manager setup
        self.file_manager = EartagFileManager(self)
        self.file_view.set_file_manager(self.file_manager)
        self.file_manager.connect("notify::is-modified", self.toggle_save_button)
        self.file_manager.connect("notify::has-error", self.toggle_save_button)
        self.file_manager.connect("notify::is-selected-modified", self.toggle_undo_all_action)
        self.file_manager.files.connect("items-changed", self.toggle_fileview)
        self.file_manager.connect("refresh-needed", self.update_state)
        self.file_manager.connect("selection-changed", self.update_state)
        self.file_manager.connect("notify::is-busy", self.update_busy)
        self.file_manager.connect("refresh-needed", self.refresh_actionbar_button_state)
        self.file_manager.files.connect("items-changed", self.refresh_actionbar_button_state)
        self.file_manager.connect("has-unwritable", self.show_unwritable_toast)
        self.file_manager.connect("selection-changed", self.refresh_actionbar_button_state)

        self.file_manager.load_task.connect("progress-pulse", self.loading_progressbar_pulse)
        self.file_manager.load_task.connect("notify::progress", self.update_loading_progressbar)
        self.file_manager.load_task.connect(
            "task-done", self.task_done_handler, EartagErrorType.ERROR_LOAD
        )
        self.file_manager.connect("max-recursion-depth-reached", self.show_max_recursion_toast)

        self.file_manager.save_task.connect("notify::progress", self.update_loading_progressbar)
        self.file_manager.save_task.connect(
            "task-done", self.task_done_handler, EartagErrorType.ERROR_SAVE
        )

        self.file_manager.rename_task.connect(
            "task-done", self.task_done_handler, EartagErrorType.ERROR_RENAME
        )

        self.show_error_action = Gio.SimpleAction.new(
            "show-error-dialog", GLib.VariantType.new("(is)")
        )
        self.show_error_action.connect("activate", self._show_error_action)
        self.add_action(self.show_error_action)

        self.connect("close-request", self.on_close_request)

        self.drop_target = Gtk.DropTarget(
            actions=Gdk.DragAction.COPY,
            formats=Gdk.ContentFormats.new_for_gtype(Gdk.FileList),
        )
        self.drop_target.connect("accept", self.on_drag_accept)
        self.drop_target.connect("enter", self.on_drag_hover)
        self.drop_target.connect("leave", self.on_drag_unhover)
        self.drop_target.connect("drop", self.on_drag_drop)
        self.add_controller(self.drop_target)

        self.toggle_fileview()

        # Tasks for undo/redo all option
        self._undo_all_count = 0
        self.undo_all_task = EartagAsyncTask(self._undo_all)
        self.undo_all_task.connect("task-done", self._undo_all_done)

        self._redo_all_count = 0
        self.redo_all_task = EartagAsyncTask(self._redo_all)
        self.redo_all_task.connect("task-done", self._redo_all_done)

        # Task for delete all tags option
        self._delete_all_tags_count = 0
        self.delete_all_tags_task = EartagAsyncTask(self._delete_all_tags)
        self.delete_all_tags_task.connect("task-done", self._delete_all_tags_done)

        self._undo_delete_all_count = 0
        self.undo_delete_all_tags_task = EartagAsyncTask(self._undo_delete_all_tags)
        self.undo_delete_all_tags_task.connect("task-done", self._undo_delete_all_tags_done)

        # Sidebar setup
        self.sidebar_file_list.set_file_manager(self.file_manager)
        self.sidebar_list_stack.set_visible_child(self.sidebar_no_files)

        self.search_bar.set_key_capture_widget(self)
        self.search_bar.connect_entry(self.search_entry)
        self.search_entry.connect("search-changed", self.search_changed)

        self.search_bar.bind_property(
            "search-mode-enabled",
            self.sidebar_search_button,
            "active",
            GObject.BindingFlags.BIDIRECTIONAL | GObject.BindingFlags.SYNC_CREATE,
        )
        self.select_multiple_button.bind_property(
            "active",
            self,
            "selection-mode",
            GObject.BindingFlags.BIDIRECTIONAL | GObject.BindingFlags.SYNC_CREATE,
        )
        self.select_multiple_button.bind_property(
            "active",
            self.sidebar_action_bar,
            "revealed",
            GObject.BindingFlags.BIDIRECTIONAL | GObject.BindingFlags.SYNC_CREATE,
        )
        self.sidebar_file_list.connect("notify::selection-mode", self.update_selection_mode)

        self.refresh_actionbar_button_state()

        if paths:
            self.file_manager.load_files(paths, overwrite=True)

    @GObject.Property(type=bool, default=False)
    def is_busy(self):
        return self.file_manager.is_busy or self.file_view.bind_in_progress

    def update_busy(self, *args):
        """
        Mark the window as busy or not busy based on whether the file manager
        is currently processing a long operation (opening/saving/renaming files).

        Also handles showing error dialogs.
        """
        self.notify("is-busy")
        busy = self.is_busy

        self.sidebar_headerbar.set_sensitive(not busy)
        self.loading_progressbar_revealer.set_reveal_child(busy)
        self.sidebar_file_list.set_sensitive(not self.file_manager.is_busy)
        self.file_view.content_clamp.set_sensitive(not busy)
        self.props.application.select_all_action.set_enabled(not busy)

        if self.file_manager.load_task.is_running:
            self.file_view.set_visible_child(self.file_view.loading)
        else:
            self.file_view.set_visible_child(self.file_view.content_stack)

        if self.file_manager.files.get_n_items() == 0:
            if busy:
                self.container_stack.set_visible_child(self.split_view)
            else:
                self.container_stack.set_visible_child(self.no_file)

    def show_error_dialog(self, error_type: EartagErrorType, error_message: str):
        """Show an error dialog with logs."""
        error_dialog = EartagErrorDialog(error_type, error_message)
        error_dialog.present(self)

    def _show_error_action(self, action: Gio.Action, data: GLib.Variant):
        """Wrapper around show_error_dialog for the win.show-error-dialog option."""
        error_type_int = data.get_child_value(0).get_int32()
        error_message = data.get_child_value(1).get_string()
        return self.show_error_dialog(EartagErrorType(error_type_int), error_message)

    def task_done_handler(self, task, error_type: EartagErrorType):
        """
        Handle a long file manager task (opening, saving or renaming) being done.

        The type of task that finished can be derived from the error_type parameter.
        """

        # If the task encountered an error, show an error dialog
        if task.errors:
            error_message = "\n---\n".join(task.errors)
            if self.file_manager.files.get_n_items() == 0:
                self.show_error_dialog(error_type, error_message)
            else:
                param = GLib.Variant.new_tuple(
                    GLib.Variant.new_int32(int(error_type)),
                    GLib.Variant.new_string(error_message),
                )
                toast = Adw.Toast.new(_("Some files failed to load"))
                toast.props.button_label = _("More Information")
                toast.props.action_name = "win.show-error-dialog"
                toast.props.action_target = param
                self.toast_overlay.add_toast(toast)

        # Handle file loading not detecting any supported files
        if error_type == EartagErrorType.ERROR_LOAD:
            if self.file_manager.load_task.n_items == 0:
                toast = Adw.Toast.new(_("No supported files found in opened folder"))
                self.toast_overlay.add_toast(toast)

        # Handle saving being done
        if error_type == EartagErrorType.ERROR_SAVE:
            saved_file_count = self.file_manager.save_task.n_items - len(
                self.file_manager.load_task.errors
            )
            save_message = ngettext(
                "Saved changes to 1 file",
                "Saved changes to {n} files",
                saved_file_count,
            ).format(n=saved_file_count)

            self.toast_overlay.add_toast(Adw.Toast.new(save_message))

    def show_unwritable_toast(self, *args):
        """Show a toast saying that some files are read-only."""
        unwritable_msg = _("Some of the opened files are read-only; changes cannot be saved")  # noqa: E501
        self.toast_overlay.add_toast(Adw.Toast.new(unwritable_msg))

    def show_max_recursion_toast(self, *args):
        """Show a toast saying that the maximum recursion depth has been reached."""
        max_depth_msg = _("Maximum folder depth reached, some files may not be loaded")
        self.toast_overlay.add_toast(Adw.Toast.new(max_depth_msg))

    def update_state(self, *args):
        app = self.get_application()

        # Set up the active view (hide fileview if there are no selected files)
        selected_files_count = self.file_manager.get_n_selected()

        if selected_files_count == self.file_manager.get_n_files():
            self.select_all_button.set_icon_name("edit-select-none-symbolic")
            self.select_all_button.set_tooltip_text(_("Unselect all files"))
        else:
            self.select_all_button.set_icon_name("edit-select-all-symbolic")
            self.select_all_button.set_tooltip_text(_("Select all files"))

        if selected_files_count <= 0:
            try:
                for action in (
                    app.rename_action,
                    app.extract_action,
                    app.identify_action,
                    app.undo_all_action,
                    app.delete_all_tags_action,
                ):
                    action.set_enabled(False)
            except AttributeError:
                return
            self.set_title(_("Ear Tag"))
            self.window_title.set_subtitle("")
            if self.file_manager.files:
                self.file_view.content_stack.set_visible_child(self.file_view.select_file)

            return False
        else:
            files = self.file_manager.selected_files_list

        self.file_view.content_stack.set_visible_child(self.file_view.content_scroll)

        # Set up window title and file info label
        _title = _("Ear Tag")
        if len(files) == 1:
            file = files[0]
            file_basename = os.path.basename(file.path)
            self.set_title("{f} — {title}".format(f=file_basename, title=_title))
            self.window_title.set_subtitle(file_basename)
        else:
            # TRANSLATORS: Placeholder for file path when multiple files are selected.
            # Shows up in the titlebar of the application.
            _multiple_files = _("(Multiple files selected)")
            self.set_title("{f} — {title}".format(f=_multiple_files, title=_title))
            self.window_title.set_subtitle(_multiple_files)

        is_modified = False
        for file in files:
            if file.is_modified:
                is_modified = True
                break

        try:
            app.undo_all_action.set_enabled(is_modified)
            for action in (
                app.rename_action,
                app.extract_action,
                app.identify_action,
                app.delete_all_tags_action,
            ):
                action.set_enabled(True)
        except AttributeError:
            pass

        self.toggle_save_button()

    def toggle_fileview(self, *args):
        """
        Shows/hides the fileview/"no files" message depending on opened files.
        """
        if self.file_manager.files.get_n_items() > 0:
            self.container_stack.set_visible_child(self.split_view)
            self.sidebar_headerbar.set_sensitive(not self.is_busy)
            self.sidebar_list_stack.set_visible_child(self.sidebar_list_scroll)
        else:
            self.container_stack.set_visible_child(self.no_file)
            self.no_file_widget.grab_button_focus()
            self.sidebar_headerbar.set_sensitive(False)
            self.sidebar_list_stack.set_visible_child(self.sidebar_no_files)
            if self.selection_mode:
                self.select_multiple_button.set_active(False)

    def on_drag_accept(self, target, drop, *args):
        drop.read_value_async(Gdk.FileList, 0, None, self.verify_files_valid)
        return True

    def verify_files_valid(self, drop, task, *args):
        try:
            files = drop.read_value_finish(task).get_files()
        except GLib.GError:
            self.drop_target.reject()
            self.on_drag_unhover()
            return False

        for file in files:
            path = file.get_path()
            if not os.path.isdir(path) and not is_valid_music_file(path):
                self.drop_target.reject()
                self.on_drag_unhover()

    def on_drag_hover(self, *args):
        self.drop_highlight_revealer.set_visible(True)
        self.drop_highlight_revealer.set_reveal_child(True)
        self.drop_highlight_revealer.set_can_target(True)
        return Gdk.DragAction.COPY

    def on_drag_unhover(self, *args):
        self.drop_highlight_revealer.set_reveal_child(False)
        self.drop_highlight_revealer.set_can_target(False)

    def on_drag_drop(self, drop_target, value, *args):
        files = value.get_files()
        self.file_manager.load_files([f.get_path() for f in files], overwrite=False)
        self.on_drag_unhover()

    @Gtk.Template.Callback()
    def drop_highlight_autohide(self, revealer, *args):
        """Sets the drop highlight to be invisible."""
        revealer.set_visible(revealer.props.child_revealed)

    async def show_file_chooser_async(self, chooser: EartagFileDialogType):
        """Show the file chooser dialog and get files from it."""

        # Set up title and check for chooser type validity
        if chooser & EartagFileDialogType.TYPE_FILE:
            if chooser & EartagFileDialogType.MODE_OVERWRITE:
                #: TRANSLATORS: Title of the popup for opening a file
                self.file_chooser.props.title = _("Open File")
            elif chooser & EartagFileDialogType.MODE_INSERT:
                self.file_chooser.props.title = _("Add File")
            else:
                raise ValueError

        elif chooser & EartagFileDialogType.TYPE_FOLDER:
            if chooser & EartagFileDialogType.MODE_OVERWRITE:
                self.file_chooser.props.title = _("Open Folder")
            else:
                raise ValueError

        else:
            raise ValueError

        # Set up filters for file selectors
        if chooser & EartagFileDialogType.TYPE_FILE:
            _filters = Gio.ListStore.new(Gtk.FileFilter)
            _filters.append(self.audio_file_filter)
            self.file_chooser.set_filters(_filters)
        else:
            self.file_chooser.set_filters(None)

        # If we're doing an overwrite load, warn about discarding changes
        if chooser & EartagFileDialogType.MODE_OVERWRITE and self.file_manager._is_modified:
            self.discard_warning = EartagDiscardWarningDialog()
            self.discard_warning.present(self)
            response = await self.discard_warning.wait_for_response()

            if response == "cancel":
                del self.discard_warning
                return
            elif response == "save":
                self.file_manager.save()

            del self.discard_warning

        # Display the dialog
        try:
            if chooser & EartagFileDialogType.TYPE_FOLDER:
                gfiles = await self.file_chooser.select_multiple_folders(self, None)
            else:
                gfiles = await self.file_chooser.open_multiple(self, None)
        except GLib.GError:
            return

        await self.file_manager.load_files_async(
            [f.get_path() for f in gfiles],
            overwrite=bool(chooser & EartagFileDialogType.MODE_OVERWRITE),
        )

    def show_file_chooser(self, chooser: EartagFileDialogType):
        """Non-async wrapper around show_file_chooser_async."""
        asyncio.create_task(self.show_file_chooser_async(chooser))

    def toggle_save_button(self, *args):
        if self.file_manager.has_error:
            self.save_button.set_tooltip_text(
                _("Some of the opened files have invalid values; cannot save")
            )  # noqa: E501
        else:
            self.save_button.set_tooltip_text("")

        self.save_button.set_sensitive(
            self.file_manager.is_modified and not self.file_manager.has_error
        )

    @Gtk.Template.Callback()
    def show_sidebar(self, *args):
        self.split_view.set_show_sidebar(True)
        self.sidebar_file_list.grab_focus()

    @Gtk.Template.Callback()
    def hide_sidebar(self, *args):
        self.split_view.set_show_sidebar(False)

    @Gtk.Template.Callback()
    def on_save(self, *args):
        self.file_manager.save()

    @Gtk.Template.Callback()
    def insert_file(self, *args):
        self.show_file_chooser(EartagFileDialogType.TYPE_FILE | EartagFileDialogType.MODE_INSERT)

    def on_close_request(self, *args):
        if (
            not self.force_close
            and list(self.file_manager.files)
            and self.file_manager._is_modified
        ):
            self.close_request_dialog = EartagCloseWarningDialog(self, self.file_manager)
            self.close_request_dialog.present(self)
            return True

        if self._identify_imported:
            from .musicbrainz import EartagCAACover

            EartagCAACover.clear_tempfiles()

        for file in self.file_manager.files:
            file.on_remove()

        self.clear_redo_data()
        self.clear_delete_all_tags_undo_data()

        self.file_view.on_close()

    def show_rename_dialog(self, *args):
        from .rename import EartagRenameDialog

        self.rename_dialog = EartagRenameDialog(self)
        self.rename_dialog.present(self)

    def show_identify_dialog(self, *args):
        from .identify import EartagIdentifyDialog

        self._identify_imported = True
        self.identify_dialog = EartagIdentifyDialog(self)
        self.identify_dialog.present(self)

    def show_preferences_dialog(self, *args):
        self.preferences_dialog = EartagPreferencesDialog()
        self.preferences_dialog.present(self)

    def show_extract_dialog(self, *args):
        from .extract import EartagExtractTagsDialog

        self.extract_dialog = EartagExtractTagsDialog(self)
        self.extract_dialog.present(self)

    def show_delete_all_tags_dialog(self, *args):
        self.delete_all_tags_dialog = EartagTagDeleteWarningDialog(self)
        self.delete_all_tags_dialog.present(self)

    # Sidebar

    def update_loading_progressbar(self, task, *args):
        """Updates the loading progressbar's position."""
        self.loading_progressbar.set_fraction(task.progress)

    def loading_progressbar_pulse(self, task, *args):
        """Updates the loading progressbar's pulse."""
        self.loading_progressbar.pulse()

    def search_changed(self, search_entry, *args):
        """Emitted when the search has changed."""
        self.file_manager.props.file_filter_str = search_entry.get_text()
        n_results = self.file_manager.file_filter_model.get_n_items()

        if n_results == 0 and self.file_manager.files.get_n_items() > 0:
            self.sidebar_list_stack.set_visible_child(self.no_results)
        else:
            self.toggle_fileview()

        if not self.selection_mode and not self.file_manager.get_n_selected() and n_results:
            self.file_manager.select_file(self.file_manager.file_filter_model[0])

        # Scroll back to top of list
        vadjust = self.sidebar_file_list.get_vadjustment()
        vadjust.set_value(vadjust.get_lower())

        # Update the switcher buttons in the fileview
        self.get_native().file_view.update_buttons()

    def toggle_selection_mode(self, *args):
        self.sidebar_file_list.toggle_selection_mode()

    @Gtk.Template.Callback()
    def select_all(self, *args):
        if self.file_manager.all_selected() and self.file_manager.get_n_selected() != 1:
            self.file_manager.unselect_all()
        else:
            self.file_manager.select_all()

    @Gtk.Template.Callback()
    def remove_selected(self, *args):
        self.sidebar_file_list.remove_selected()

    def refresh_actionbar_button_state(self, *args):
        if not self.file_manager.files or not self.selection_mode:
            selected_message = ""
            self.sidebar_action_bar.set_sensitive(False)
            self.sidebar_action_bar.set_revealed(False)
        else:
            self.sidebar_action_bar.set_sensitive(True)
            self.sidebar_action_bar.set_revealed(True)
            selected_file_count = self.file_manager.get_n_selected()
            if selected_file_count == 0:
                selected_message = _("No files selected")
                self.remove_selected_button.set_sensitive(False)
            else:
                # TRANSLATORS: {n} is a placeholder for the amount of files.
                # **Do not change the letter between the curly brackets!**
                selected_message = ngettext(
                    "1 file selected", "{n} files selected", selected_file_count
                ).format(n=selected_file_count)
                self.remove_selected_button.set_sensitive(True)

        self.selected_message_label.set_label(selected_message)

    @GObject.Property(type=bool, default=False)
    def selection_mode(self):
        """Whether the sidebar is in selection mode or not."""
        return self.sidebar_file_list.selection_mode

    @selection_mode.setter
    def selection_mode(self, value):
        self.sidebar_file_list.selection_mode = value
        # Workaround for the text not showing up on initial load
        self.refresh_actionbar_button_state()

    def update_selection_mode(self, filelist, *args):
        self.notify("selection-mode")
        # Workaround for the text not showing up on initial load
        self.refresh_actionbar_button_state()

    def select_next(self, *args):
        """Selects the next item on the sidebar."""
        return self.file_manager.select_next()

    def select_previous(self, *args):
        """Selects the previous item on the sidebar."""
        return self.file_manager.select_previous()

    def scroll_to_top(self):
        """Scrolls to the top of the file list."""
        self.sidebar_list_scroll.get_vadjustment().set_value(0)

    def scroll_to_index(self, index):
        """Scrolls to file at the specified index."""
        item_height = self.sidebar_file_list.get_first_child().get_height()
        vadjust = self.sidebar_list_scroll.get_vadjustment()
        new_value = item_height * (index + 1)
        if new_value <= vadjust.get_upper():
            vadjust.set_value(new_value)
        else:
            vadjust.set_value(vadjust.get_upper())

    # Undo all option

    def undo_all(self, *args):
        self.set_sensitive(False)
        self.undo_all_task.run()

    async def _undo_all(self, *args):
        """Undo all changes and add the option to re-do them."""
        self._undo_all_data = {}
        self._undo_all_count = 0
        for file in self.file_manager.selected_files_list:
            if not file.is_modified:
                continue
            self._undo_all_count += 1

            self._undo_all_data[file.id] = {}
            for tag in file.modified_tags:
                self._undo_all_data[file.id][tag] = file.get_property(tag)

            file.undo_all()

    def _undo_all_done(self, *args):
        self._redo_all_called = False
        self.set_sensitive(True)

        toast = Adw.Toast.new(
            # TRANSLATORS: {n} is a placeholder for the amount of files.
            # **Do not change the letter between the curly brackets!**
            ngettext(
                "Undid changes in 1 file",
                "Undid changes in {n} files",
                self._undo_all_count,
            ).format(n=self._undo_all_count)
        )
        toast.props.button_label = _("Redo")
        toast.connect("button-clicked", self.redo_all)
        toast.connect("dismissed", self.clear_redo_data)
        self.toast_overlay.add_toast(toast)

    def redo_all(self, *args):
        self._redo_all_called = True
        self.set_sensitive(False)
        self.redo_all_task.run()

    async def _redo_all(self, *args):
        """Reverses undo_all."""
        self._redo_all_count = 0
        for file in self.file_manager.files:
            if file.id in self._undo_all_data:
                self._redo_all_count += 1
                for tag, value in self._undo_all_data[file.id].items():
                    if tag in file.int_properties + file.float_properties and not value:
                        file.set_property(tag, 0)
                    else:
                        file.set_property(tag, value)

    def _redo_all_done(self, *args):
        self.set_sensitive(True)
        toast = Adw.Toast.new(
            # TRANSLATORS: {n} is a placeholder for the amount of files.
            # **Do not change the letter between the curly brackets!**
            ngettext(
                "Redid changes in 1 file",
                "Redid changes in {n} files",
                self._redo_all_count,
            ).format(n=self._redo_all_count)
        )
        self.toast_overlay.add_toast(toast)
        self._redo_all_called = False
        self.clear_redo_data()

    def clear_redo_data(self, *args):
        """Clears the redo data for the "undo all" option."""
        if not self._redo_all_called:
            del self._undo_all_data
            self._undo_all_data = {}

    def toggle_undo_all_action(self, *args):
        app = self.get_application()
        try:
            app.undo_all_action.set_enabled(self.file_manager.props.is_selected_modified)
        except AttributeError:
            pass

    # "Delete all tags" option
    def do_delete_all_tags(self, *args):
        self.set_sensitive(False)
        self.delete_all_tags_task.run()

    async def _delete_all_tags(self, *args):
        """Undo all changes and add the option to re-do them."""
        self._delete_all_tags_undo_data = {}
        files = self.file_manager.selected_files_list.copy()
        self._delete_all_tags_count = len(files)
        for file in files:
            self._delete_all_tags_undo_data[file.id] = {}
            for prop in file.present_tags:
                self._delete_all_tags_undo_data[file.id][prop] = file.get_property(prop)
            file.delete_all_raw()

    def _delete_all_tags_done(self, *args):
        self.set_sensitive(True)

        toast = Adw.Toast.new(
            # TRANSLATORS: {n} is a placeholder for the amount of files
            # that tags have been removed from.
            # **Do not change the letter between the curly brackets!**
            ngettext(
                "Removed tags from 1 file",
                "Removed tags from {n} files",
                self._delete_all_tags_count,
            ).format(n=self._delete_all_tags_count)
        )
        toast.props.button_label = _("Undo")
        toast.connect("button-clicked", self.undo_delete_all_tags)
        # toast.connect('dismissed', self.clear_delete_all_tags_undo_data)
        self.toast_overlay.add_toast(toast)

    def undo_delete_all_tags(self, *args):
        self.set_sensitive(False)
        self.undo_delete_all_tags_task.run()

    async def _undo_delete_all_tags(self, *args):
        """Reverses undo_all."""
        self._undo_delete_all_count = 0
        for file in self.file_manager.files:
            if file.id in self._delete_all_tags_undo_data:
                self._undo_delete_all_count += 1
                await file.reload()
                for prop, value in self._delete_all_tags_undo_data[file.id].items():
                    file.set_property(prop, value)

    def _undo_delete_all_tags_done(self, *args):
        self.set_sensitive(True)

        toast = Adw.Toast.new(
            # TRANSLATORS: {n} is a placeholder for the amount of files.
            # **Do not change the letter between the curly brackets!**
            ngettext(
                "Undid tag removal in 1 file",
                "Undid tag removal in {n} files",
                self._undo_delete_all_count,
            ).format(n=self._undo_delete_all_count)
        )
        self.toast_overlay.add_toast(toast)

    def clear_delete_all_tags_undo_data(self, *args):
        self._delete_all_tags_undo_data = {}


@Gtk.Template(resource_path=f"{APP_GRESOURCE_PATH}/ui/settings.ui")
class EartagPreferencesDialog(Adw.PreferencesDialog):
    __gtype_name__ = "EartagPreferencesDialog"

    open_folders_recursively_toggle = Gtk.Template.Child()
    mb_confidence_spinbutton = Gtk.Template.Child()
    aid_confidence_spinbutton = Gtk.Template.Child()
    cover_size_comborow = Gtk.Template.Child()

    def __init__(self):
        super().__init__()

        config.bind(
            "open-folders-recursively",
            self.open_folders_recursively_toggle,
            "active",
            flags=Gio.SettingsBindFlags.DEFAULT,
        )

        config.bind(
            "musicbrainz-confidence-treshold",
            self.mb_confidence_spinbutton,
            "value",
            flags=Gio.SettingsBindFlags.DEFAULT,
        )
        config.bind(
            "acoustid-confidence-treshold",
            self.aid_confidence_spinbutton,
            "value",
            flags=Gio.SettingsBindFlags.DEFAULT,
        )

        self.bind_property(
            "cover-size-setting",
            self.cover_size_comborow,
            "selected",
            GObject.BindingFlags.BIDIRECTIONAL | GObject.BindingFlags.SYNC_CREATE,
        )

    @GObject.Property(type=int)
    def cover_size_setting(self):
        if config.get_enum("musicbrainz-cover-size") == 0:
            return 0
        elif config.get_enum("musicbrainz-cover-size") == 250:
            return 1
        elif config.get_enum("musicbrainz-cover-size") == 500:
            return 2
        elif config.get_enum("musicbrainz-cover-size") == 1200:
            return 3
        return 4

    @cover_size_setting.setter
    def cover_size_setting(self, value):
        config.set_enum("musicbrainz-cover-size", int(DLCoverSize.index_to_item(value)))



# SPDX-License-Identifier: MIT
# (c) 2023 knuxify and Ear Tag contributors
"""
Common code for the placeholder selector used in the rename and extract UIs,
as well as the fileview "add extra tags" button.
"""

from .. import APP_GRESOURCE_PATH
from ..backends.file import VALID_TAGS, TAG_NAMES

from gi.repository import Gdk, Gtk, GObject


@Gtk.Template(resource_path=f"{APP_GRESOURCE_PATH}/ui/tagselectorbutton.ui")
class EartagTagSelectorButton(Gtk.MenuButton):
    """
    Button that allows you to select a tag from a list.

    Users can bind to the tag-selected signal to get the name
    of the tag.
    """

    __gtype_name__ = "EartagTagSelectorButton"

    tag_list_popover = Gtk.Template.Child()
    tag_list = Gtk.Template.Child()
    tag_list_search_entry = Gtk.Template.Child()

    def __init__(self):
        # Extra tag filter for additional tag field
        self.tag_names = dict(
            [(k, v) for k, v in TAG_NAMES.items() if k in VALID_TAGS + ("length", "bitrate")]
        )
        self.tag_names_swapped = dict([(v, k) for k, v in self.tag_names.items()])
        tag_model_nofilter = Gtk.StringList.new(list(self.tag_names.values()))
        self.tag_model = Gtk.FilterListModel(model=tag_model_nofilter)
        self.tag_filter = Gtk.CustomFilter.new(self.tag_filter_func, self.tag_model)
        self.tag_model.set_filter(self.tag_filter)

        self.tag_selection_model = Gtk.NoSelection.new(self.tag_model)

        # "Add additional tag" field (we can reuse the factory from the "more tags" selector):
        self._ignore_tag_selector = False

        factory = Gtk.BuilderListItemFactory.new_from_resource(
            None, f"{APP_GRESOURCE_PATH}/ui/moretagsgroupfactory.ui"
        )
        self.tag_list.set_model(self.tag_selection_model)
        self.tag_list.set_factory(factory)
        self.tag_list.connect("activate", self.add_placeholder_from_selector)

        # Close popover if Escape key is pressed in search entry
        controller = Gtk.ShortcutController()
        trigger = Gtk.KeyvalTrigger.new(Gdk.keyval_from_name("Escape"), 0)
        shortcut = Gtk.Shortcut.new(trigger, Gtk.CallbackAction.new(self.close_popover))
        controller.add_shortcut(shortcut)
        self.tag_list_search_entry.add_controller(controller)

    def add_placeholder_from_selector(self, listview, position, *args):
        """Adds a new placeholder based on the tag selector."""
        if self._ignore_tag_selector:
            return

        self._ignore_tag_selector = True

        selected_item = self.tag_selection_model.get_item(position)
        if not selected_item:
            return
        if selected_item.get_string() == "none":
            return
        tag = self.tag_names_swapped[selected_item.get_string()]

        self.tag_list_popover.popdown()

        self.emit("tag-selected", tag)

        self._ignore_tag_selector = False

    @GObject.Signal(arg_types=(str,))
    def tag_selected(self, tag: str):
        """Emitted when a tag is selected."""

    def tag_filter_func(self, _tag_name, *args):
        """Filter function for the tag dropdown."""
        tag_name = _tag_name.get_string()
        query = self.get_search_query()
        if query:
            return query.lower() in tag_name.lower()
        return True

    def close_popover(self, *args):
        self.tag_list_popover.popdown()

    @Gtk.Template.Callback()
    def refresh_tag_filter(self, *args):
        """Refreshes the filter for the tag placeholder insert row."""
        self.tag_filter.changed(Gtk.FilterChange.DIFFERENT)

    def set_filter(self, filter):
        """
        Sets a custom filter for the tag selector.

        The primary user of this function is the FileView code, which needs
        to manage its own filter model to keep track of available tags and
        remove already present ones.
        """
        self.tag_filter = filter
        self.tag_model.set_filter(self.tag_filter)

    def get_search_query(self) -> str:
        """Returns the search query typed into the search entry."""
        return self.tag_list_search_entry.get_text()



# SPDX-License-Identifier: MIT
# (c) 2023 knuxify and Ear Tag contributors

import gi

gi.require_version("GdkPixbuf", "2.0")
from gi.repository import GObject, GdkPixbuf, GLib, Gio
import os
import re
import shutil
import uuid
import asyncio
import aiofiles
import xxhash
from PIL import Image
from typing import Self
from enum import IntEnum
import traceback
import io
import weakref

from ..utils.validation import get_mimetype, get_mimetype_buffer
from ..utils.misc import safe_int, safe_float
from ..logger import logger

BASIC_TAGS = (
    "title", "artist", "album", "albumartist", "tracknumber",
    "totaltracknumber", "genre", "releasedate", "comment"
)  # fmt: skip

EXTRA_TAGS = (
    "bpm", "compilation", "composer", "copyright", "encodedby",
    "mood", "conductor", "arranger", "discnumber", "publisher",
    "isrc", "language", "discsubtitle", "url",

    "albumartistsort", "albumsort", "composersort", "artistsort",
    "titlesort",

    "musicbrainz_artistid", "musicbrainz_albumid",
    "musicbrainz_albumartistid", "musicbrainz_trackid",
    "musicbrainz_recordingid", "musicbrainz_releasegroupid"
)  # fmt: skip

VALID_TAGS = BASIC_TAGS + EXTRA_TAGS


class CoverType(IntEnum):
    FRONT = 0
    BACK = 1


# Workaround for tests not having the _ variable available
try:
    _  # noqa: B018
except NameError:
    _ = lambda x: x  # noqa: E731

# Human-readable tag names
TAG_NAMES = {
    "length": _("Length"),
    "bitrate": _("Bitrate"),

    "title": _("Title"),
    "artist": _("Artist"),
    "album": _("Album"),
    "albumartist": _("Album artist"),
    "tracknumber": _("Track number"),
    "totaltracknumber": _("Total tracks"),
    "genre": _("Genre"),
    "releasedate": _("Release date"),
    "comment": _("Comment"),

    "none": _("(Select a tag)"),
    # TRANSLATORS: Short for "beats per minute".
    "bpm": _("BPM"),
    "compilation": _("Compilation"),
    "composer": _("Composer"),
    "copyright": _("Copyright"),
    "encodedby": _("Encoded by"),
    "mood": _("Mood"),
    # TRANSLATORS: Orchestra conductor
    "conductor": _("Conductor"),
    "arranger": _("Arranger"),
    "discnumber": _("Disc number"),
    "publisher": _("Publisher"),
    "isrc": "ISRC",
    "language": _("Language"),
    "discsubtitle": _("Disc subtitle"),
    "url": _("Website/URL"),

    # TRANSLATORS: This is a sort tag, as in, a tag that dictates how music
    # software should treat this tag when sorting.
    "albumartistsort": _("Album artist (sort)"),
    # TRANSLATORS: This is a sort tag, as in, a tag that dictates how music
    # software should treat this tag when sorting.
    "albumsort": _("Album (sort)"),
    # TRANSLATORS: This is a sort tag, as in, a tag that dictates how music
    # software should treat this tag when sorting.
    "composersort": _("Composer (sort)"),
    # TRANSLATORS: This is a sort tag, as in, a tag that dictates how music
    # software should treat this tag when sorting.
    "artistsort": _("Artist (sort)"),
    # TRANSLATORS: This is a sort tag, as in, a tag that dictates how music
    # software should treat this tag when sorting.
    "titlesort": _("Title (sort)"),

    "musicbrainz_artistid": _("MusicBrainz Artist ID"),
    "musicbrainz_albumid": _("MusicBrainz Album ID"),
    "musicbrainz_albumartistid": _("MusicBrainz Album Artist ID"),
    "musicbrainz_trackid": _("MusicBrainz Release Track ID"),
    "musicbrainz_recordingid": _("MusicBrainz Recording ID"),
    "musicbrainz_releasegroupid": _("MusicBrainz Release Group ID"),
}  # fmt: skip


class EartagFileCover:
    """
    Represents a single cover image. Used to store pixbufs for thumbnails
    and for cover comparisons.
    """

    #: Cover cache for cover object reuse. The key is the hash of the file, and the value is a list containing the current refcount and the EartagFileCover object.
    _cache: dict[bytes, Self] = weakref.WeakValueDictionary()

    def __init__(self):
        """Do not call this function directly, use new_from_bytes or new_from_path instead."""

        #: Cover path, if this cover was loaded from a file.
        self.path: str | None = None

        #: Cover hash digest as generated by xxhash.
        self.hash: bytes | None = None

        #: Cover file data as loaded from the file.
        self.data: bytes | None = None

        #: Mimetype of the cover, as a string.
        self.mime: str | None = None

    def __del__(self):
        del self.data
        if self.hash and self.hash in self._cache:
            del self._cache[self.hash]

    @classmethod
    async def new_from_bytes(cls, data: bytes) -> Self:
        """
        Create a new EartagFileCover (or get a cached one) based on the cover data bytes.
        """
        _hash = await asyncio.to_thread(xxhash.xxh3_64, data)
        cover_hash = _hash.digest()

        if cover_hash in cls._cache:
            return cls._cache[cover_hash]

        mime = get_mimetype_buffer(data)
        if not mime or not mime.startswith("image/"):
            return None

        out = cls()
        out.mime = mime
        out.hash = cover_hash
        out.data = data

        try:
            # Load using GdkPixbuf
            async def _get_pixbuf_data(data: bytes, size: int):
                data_bytes = GLib.Bytes.new(data)
                data_stream = Gio.MemoryInputStream.new_from_bytes(data_bytes)

                out = GdkPixbuf.Pixbuf.new_from_stream_at_scale(data_stream, size, size, True)

                data_stream.close()
                del data_stream
                data_bytes.unref()
                del data_bytes

                return out

            async with asyncio.TaskGroup() as tg:
                task_small = tg.create_task(_get_pixbuf_data(data, 48))
                task_large = tg.create_task(_get_pixbuf_data(data, 196))
            out.cover_small = task_small.result()
            out.cover_large = task_large.result()

        except GLib.GError:
            logger.warning("An error occured while loading the image with GdkPixbuf:")
            traceback.print_exc()
            logger.warning("Falling back to loading with PIL.")

            try:
                img = await asyncio.to_thread(Image.open, io.BytesIO(data))
                if img.mode != "RGB":
                    await asyncio.to_thread(img.convert, "RGB")

                async def _get_pixbuf_pil(img: Image) -> GdkPixbuf.Pixbuf:
                    data = bytes(x for xs in img.getdata() for x in xs)

                    return await asyncio.to_thread(
                        GdkPixbuf.Pixbuf.new_from_data,
                        data,
                        GdkPixbuf.Colorspace.RGB,
                        img.has_transparency_data,
                        8,
                        img.width,
                        img.height,
                        img.width * 3,
                    )

                await asyncio.to_thread(img.thumbnail, (196, 196))
                out.cover_large = await _get_pixbuf_pil(img)

                await asyncio.to_thread(img.thumbnail, (48, 48))
                out.cover_small = await _get_pixbuf_pil(img)

                img.close()

            except:  # noqa: E722
                logger.error("Failed to load cover image:")
                traceback.print_exc()
                out = None

        cls._cache[cover_hash] = out

        return out

    @classmethod
    async def new_from_path(cls, cover_path: str) -> Self:
        """
        Create a new EartagFileCover (or get a cached one) based on the
        contents of a file.
        """
        async with aiofiles.open(cover_path, "rb") as cover_file:
            data = await cover_file.read()

        ret = await cls.new_from_bytes(data)
        if not ret:
            return None

        ret.path = cover_path
        return ret

    async def save_to_path(self, target_path: str):
        """Save the data from this cover to a file."""
        if not self.data:
            return

        async with aiofiles.open(target_path, "wb") as cover_file:
            await cover_file.write(self.data)

    def __eq__(self, other):
        if not isinstance(other, EartagFileCover):
            return False
        return self.hash == other.hash


class EartagFile(GObject.Object):
    """
    Generic base for GObject wrappers that provide information about a music
    file.

    The following functions are implemented by the subclasses:
      - save() - saves the changes to a file.
      - get_tag(tag_name) - gets the tag with the given name. The name matches
                            the property name of the tag in this object.
      - set_tag(tag_name, value) - gets the tag with the given name. The name
                           matches the property name of the tag in this object.

    Do not use get_tag and set_tag directly in the code; use get_property and
    set_property instead.
    """

    __gtype_name__ = "EartagFile"

    handled_properties = BASIC_TAGS
    int_properties = ("tracknumber", "totaltracknumber", "discnumber")
    float_properties = ("bpm",)
    supported_extra_tags = []
    _supports_album_covers = False
    _is_modified = False
    _is_writable = False
    _has_error = False

    def __init__(self, path):
        """Initializes an EartagFile for the given file path."""
        super().__init__()
        self.notify("supports-album-covers")
        self._path = path
        self.update_writability()
        self.modified_tags = []
        self.original_values = {}
        self._error_fields = []
        self._releasedate_cached = None
        self._no_cover_notify = False

        self.id = str(uuid.uuid4())  # Internal ID used for keeping track of files

        #: Dict with covers belonging to this file.
        self.covers: dict[CoverType, EartagFileCover | None] = {}

    @classmethod
    async def new_from_path(cls, path):
        """Create a new EartagFile instance for the file with the given path."""
        file = cls(path)
        await file.load_from_file(path)
        return file

    def on_remove(self, *args):
        del self.modified_tags
        del self.original_values
        _cover_types = list(self.covers.keys())
        for cover_type in _cover_types:
            del self.covers[cover_type]

    def save(self):
        """Saves tag changes to the file. Must be implemented by the backend."""
        raise NotImplementedError

    def get_tag(self, tag_name):
        """Gets a tag for a specific property. Must be implemented by the backend."""
        raise NotImplementedError

    def set_tag(self, tag_name, value):
        """Sets a tag for a specific property. Must be implemented by the backend."""
        raise NotImplementedError

    def has_tag(self, tag_name):
        """Returns whether a tag is present in a file. Must be implemented by the backend."""
        raise NotImplementedError

    def delete_tag(self, tag_name):
        """Deletes the tag from the file. Must be implemented by the backend."""
        raise NotImplementedError

    def delete_all_raw(self):
        """
        Completely removes all metadata, including information that Ear Tag cannot
        read. Implementations must call the regular delete_all function first.
        This is used by the "clear all tags" option.

        Must be implemented by the backend.
        """
        raise NotImplementedError

    def delete_all(self):
        """Deletes all present tags."""
        for tag in self.present_tags:
            self.delete_tag(tag)
        for cover_type in (CoverType.FRONT, CoverType.BACK):
            self.delete_cover(cover_type)

    @property
    def present_tags(self) -> list:
        """Returns a list of present tags."""
        present_tags = []
        for tag in self.handled_properties:
            if self.has_tag(tag):
                present_tags.append(tag)
        return present_tags + self.present_extra_tags

    def setup_present_extra_tags(self):
        """
        For performance reasons, each file keeps a list of present extra tags.
        This is used by the fileview to determine which rows to display for
        extra tags.
        """
        self.present_extra_tags = []

        for tag in self.supported_extra_tags:
            if self.has_tag(tag):
                self.present_extra_tags.append(tag)

    def setup_original_values(self):
        """
        Sets up the list of original values to compare new values against.
        Run this every time the file is loaded or saved.

        Call this only after setup_present_extra_tags and load_cover have been called.
        """
        self.original_values = {}
        for tag in set(tuple(self.handled_properties) + tuple(self.present_extra_tags)):
            value = self.get_property(tag)
            if tag in self.int_properties + self.float_properties and value is None:
                self.original_values[tag] = 0
            else:
                self.original_values[tag] = self.get_property(tag)
        if self._supports_album_covers:
            self.original_values["front_cover"] = self.get_property("front_cover")
            self.original_values["back_cover"] = self.get_property("back_cover")

    def update_writability(self):
        """
        Checks if the file is writable and sets the writable property
        accordingly.
        """
        try:
            writable_check = open(self.path, "a")
            writable_check.close()
        except OSError:
            self._is_writable = False
        else:
            self._is_writable = True
        self.notify("is_writable")

    def set_error(self, field, has_error):
        """Sets an error for the given field."""
        if not has_error and field in self._error_fields:
            self._error_fields.remove(field)
        elif has_error and field not in self._error_fields:
            self._error_fields.append(field)

        self._has_error = has_error
        self.notify("has-error")

    @GObject.Signal(arg_types=(str,))
    def modified(self, tag):
        if not tag:
            return

        new_value = self.get_property(tag)
        if tag in self.supported_extra_tags:
            if new_value and tag not in self.present_extra_tags:
                self.present_extra_tags.append(tag)
            elif not new_value and tag in self.present_extra_tags:
                self.present_extra_tags.remove(tag)

        if tag in self.original_values:
            old_value = self.original_values[tag]
            if old_value == new_value or (not old_value and not new_value):
                if tag in self.modified_tags:
                    self.modified_tags.remove(tag)
            elif tag not in self.modified_tags:
                self.modified_tags.append(tag)
        elif not new_value:
            if tag in self.modified_tags:
                self.modified_tags.remove(tag)
        elif tag not in self.modified_tags:
            self.modified_tags.append(tag)

        if not self.modified_tags:
            self._is_modified = False
            self.notify("is_modified")

    @GObject.Property(type=str)
    def path(self):
        """Full path to the file."""
        return self._path

    @path.setter
    def path(self, value):
        """Moves the file to the new location."""
        raise ValueError("Use set_path_async to change the file path.")

    async def set_path_async(self, value: str):
        """
        Moves the file to the new location.
        (Internal function that does not notify the path property; used by the
        rename UI to prevent crashes.)
        """
        if value == self._path:
            return

        modifications = {}
        for tag in self.modified_tags:
            modifications[tag] = self.get_property(tag)

        await asyncio.to_thread(shutil.move, self._path, value)
        self._path = value
        await self.load_from_file(value)

        for tag, tag_value in modifications.items():
            self.set_property(tag, tag_value)

    async def reload(self):
        """Reloads the file and discards all modifications."""
        await self.load_from_file(self.props.path)
        for prop in BASIC_TAGS + tuple(self.supported_extra_tags) + ("front_cover", "back_cover"):
            self.notify(prop)

        self.mark_as_unmodified()

    def mark_as_modified(self, tag, notify_prop: bool = False):
        if not self._is_modified:
            self._is_modified = True
            self.notify("is_modified")
        if notify_prop:
            self.notify(tag)
        self.emit("modified", tag)

    def mark_as_unmodified(self):
        if self._is_modified:
            self._is_modified = False
            self.notify("is_modified")
        self.emit("modified", None)
        self.modified_tags = []

    def mark_tag_as_unmodified(self, tag):
        was_modified = bool(self.modified_tags)
        try:
            self.modified_tags.remove(tag)
        except ValueError:
            return
        if bool(self.modified_tags) != was_modified:
            self._is_modified = bool(self.modified_tags)
            self.notify("is_modified")
        self.emit("modified", None)

    def undo_all(self):
        """Undo all changes."""
        for tag in self.modified_tags.copy():
            if tag in self.original_values:
                print(tag)
                self.set_property(tag, self.original_values[tag])
            else:
                self.delete_tag(tag)
        self.mark_as_unmodified()

    def reset_to_original(self):
        """Resets to original values."""
        for tag, value in self.original_values.items():
            self.set_property(tag, value)
        self.mark_as_unmodified()

    @GObject.Property(type=bool, default=False)
    def is_modified(self):
        """Returns whether the values have been modified or not."""
        return self._is_modified

    @GObject.Property(type=bool, default=False)
    def has_error(self):
        """Returns whether or not there's an error with the file's values."""
        return self._has_error

    @GObject.Property(type=bool, default=False)
    def supports_album_covers(self):
        """Returns whether album covers are supported."""
        return self._supports_album_covers

    @GObject.Property(type=bool, default=False)
    def is_writable(self):
        """Returns whether the file can be written to."""
        return self._is_writable

    # Cover art handling functions

    def get_cover(self, cover_type: CoverType | int):
        """Gets a cover for the given cover type."""
        return self.covers.get(int(cover_type), None)

    @GObject.Property(default=None)
    def front_cover(self) -> EartagFileCover | None:
        return self.get_cover(CoverType.FRONT)

    @front_cover.setter
    def front_cover(self, value: EartagFileCover | None):
        if not value:
            self.delete_cover(CoverType.FRONT)
        else:
            self.covers[CoverType.FRONT] = value
            self._no_cover_notify = True
            self.emit("cover-updated", CoverType.FRONT)
            self._no_cover_notify = False

    @GObject.Property(default=None)
    def back_cover(self) -> EartagFileCover | None:
        return self.get_cover(CoverType.BACK)

    @back_cover.setter
    def back_cover(self, value):
        if not value:
            self.delete_cover(CoverType.BACK)
        else:
            self.covers[CoverType.BACK] = value
            self._no_cover_notify = True
            self.emit("cover-updated", CoverType.BACK)
            self._no_cover_notify = False

    def has_cover(self, cover_type: CoverType | int):
        """Check whether a cover with the given type is present in the file."""
        return self.get_cover(cover_type) is not None

    @GObject.Signal(arg_types=(int,))
    def cover_updated(self, cover_type: CoverType | int):
        """Emitted when a cover is updated; the argument is the cover type."""
        if self._no_cover_notify:
            return
        if cover_type == CoverType.FRONT:
            self.notify("front-cover")
        elif cover_type == CoverType.BACK:
            self.notify("back-cover")

    async def set_cover_from_data(
        self, cover_type: CoverType | int, data: bytes, mime: str | None = None
    ):
        raise NotImplementedError

    async def set_cover_from_path(self, cover_type: CoverType | int, path: str | None):
        if not path:
            self.delete_cover()
            return

        data, mime = await self._load_and_convert_image(path, self._cover_mimetypes)
        if not data:
            return

        await self.set_cover_from_data(cover_type, data, mime=mime)

    async def _set_cover_from_data(
        self, cover_type: CoverType | int, data: bytes, modified: bool = True
    ) -> bool:
        """
        Update the cover data with the given data bytes.

        Returns True if the cover loaded succesfully, False otherwise.
        """
        _cover_type = int(cover_type)

        # Remove conflicting covers
        self.delete_cover(cover_type, clear_only=True)

        self.covers[_cover_type] = await EartagFileCover.new_from_bytes(data)
        self.emit("cover-updated", _cover_type)

        # Mark as modified
        if modified:
            if cover_type == CoverType.FRONT:
                self.mark_as_modified("front_cover")
            elif cover_type == CoverType.BACK:
                self.mark_as_modified("back_cover")

        # If the cover failed to load, its value will be None
        return not not self.covers[_cover_type]

    async def _load_and_convert_image(
        self, path: str, allowed_mimetypes: list[str] | None = None
    ) -> (bytes | None, str | None):
        """
        Helper function for cover implementations to load an image file and check
        its mimetype, converting if needed.
        """
        mime = get_mimetype(path)
        if not mime or not mime.startswith("image/"):
            return (None, None)

        if not allowed_mimetypes or mime in allowed_mimetypes:
            async with aiofiles.open(path, "rb") as cover_file:
                data = await cover_file.read()
        else:
            # Convert to PNG
            def _convert(path):
                out = io.BytesIO()
                with Image.open(path) as img:
                    img.save(out, format="PNG")
                return out.getvalue()

            try:
                data = await asyncio.to_thread(_convert, path)
            except:  # noqa: E722
                logger.error("Failed to convert image to PNG:")
                traceback.print_exc()
                data = None
            else:
                mime = "image/png"

        return data, mime

    def _cleanup_cover(self, cover_type: CoverType | int):
        """Remove the cover with the given type from the covers dict."""
        _cover_type = int(cover_type)
        if _cover_type in self.covers:
            del self.covers[_cover_type]
        self.emit("cover-updated", _cover_type)
        if cover_type == CoverType.FRONT:
            self.mark_as_modified("front_cover")
        elif cover_type == CoverType.BACK:
            self.mark_as_modified("back_cover")

    # Properties, used for bindings; backends must implement get_tag and set_tag options
    # that take these property names, or redefine the properties

    @GObject.Property(type=str, flags=GObject.ParamFlags.READABLE)
    def filetype(self):
        return os.path.splitext(self.path)[-1]

    @GObject.Property(type=str)
    def title(self):
        return self.get_tag("title")

    @title.setter
    def title(self, value):
        if value:
            self.set_tag("title", value)
            self.mark_as_modified("title")
        elif self.has_tag("title"):
            self.delete_tag("title")

    @GObject.Property(type=str)
    def artist(self):
        return self.get_tag("artist")

    @artist.setter
    def artist(self, value):
        if value:
            self.set_tag("artist", value)
            self.mark_as_modified("artist")
        elif self.has_tag("artist"):
            self.delete_tag("artist")

    @GObject.Property(type=int)
    def tracknumber(self):
        value = self.get_tag("tracknumber")
        if value:
            return safe_int(value)
        return None

    @tracknumber.setter
    def tracknumber(self, value):
        if value:
            self.set_tag("tracknumber", safe_int(value))
            self.mark_as_modified("tracknumber")
        else:
            self.delete_tag("tracknumber")

    @GObject.Property(type=int)
    def totaltracknumber(self):
        value = self.get_tag("totaltracknumber")
        if value:
            return safe_int(value)
        return None

    @totaltracknumber.setter
    def totaltracknumber(self, value):
        if value:
            self.set_tag("totaltracknumber", safe_int(value))
            self.mark_as_modified("totaltracknumber")
        else:
            self.delete_tag("totaltracknumber")

    @GObject.Property(type=str)
    def album(self):
        return self.get_tag("album")

    @album.setter
    def album(self, value):
        if value:
            self.set_tag("album", value)
            self.mark_as_modified("album")
        elif self.has_tag("album"):
            self.delete_tag("album")

    @GObject.Property(type=str)
    def albumartist(self):
        return self.get_tag("albumartist")

    @albumartist.setter
    def albumartist(self, value):
        if value:
            self.set_tag("albumartist", value)
            self.mark_as_modified("albumartist")
        elif self.has_tag("albumartist"):
            self.delete_tag("albumartist")

    @GObject.Property(type=str)
    def genre(self):
        return self.get_tag("genre")

    @genre.setter
    def genre(self, value):
        if value:
            self.set_tag("genre", value)
            self.mark_as_modified("genre")
        elif self.has_tag("genre"):
            self.delete_tag("genre")

    # Release dates have custom handling, as invalid values don't get
    # saved correctly, so we only save valid ones to the file itself,
    # and the rest it stored internally:

    def validate_date(self, field, value):
        if not value:
            self.set_error(field, False)
            return

        has_error = True
        if "-" in value:
            for format in (
                "^[0-9]{4}$",
                "^[0-9]{4}-[0-9]{2}$",
                "^[0-9]{4}-[0-9]{2}-[0-9]{2}$",
            ):
                if re.match(format, value):
                    has_error = False
                    break
        else:
            has_error = False

        self.set_error(field, has_error)

    @GObject.Property(type=str)
    def releasedate(self):
        if not self._releasedate_cached and self.has_tag("releasedate"):
            value = self.get_tag("releasedate")
            if value and len(value) > 10:
                value = value[:10]
            self._releasedate_cached = value
        return self._releasedate_cached

    @releasedate.setter
    def releasedate(self, value):
        self.validate_date("releasedate", value)
        self._releasedate_cached = value
        if "releasedate" not in self._error_fields:
            if value:
                self.set_tag("releasedate", value)
            elif self.has_tag("releasedate"):
                self._releasedate_cached = None
                self.delete_tag("releasedate")
        self.mark_as_modified("releasedate")

    @GObject.Property(type=str)
    def comment(self):
        return self.get_tag("comment")

    @comment.setter
    def comment(self, value):
        if value:
            self.set_tag("comment", value)
            self.mark_as_modified("comment")
        elif self.has_tag("comment"):
            self.delete_tag("comment")

    # Additional tag properties.

    @GObject.Property(type=float)
    def bpm(self):
        if "bpm" in self.supported_extra_tags:
            value = self.get_tag("bpm")
            if value:
                # Some BPMs can be floating point values, so we treat this as a float
                try:
                    return safe_float(self.get_tag("bpm"))
                except ValueError:
                    return None
        return None

    @bpm.setter
    def bpm(self, value):
        if "bpm" not in self.supported_extra_tags:
            return None
        if value:
            self.set_tag("bpm", safe_float(value))
            self.mark_as_modified("bpm")
        elif self.has_tag("bpm"):
            self.delete_tag("bpm")

    @GObject.Property(type=str)
    def compilation(self):
        if "compilation" in self.supported_extra_tags:
            return self.get_tag("compilation")
        return None

    @compilation.setter
    def compilation(self, value):
        if "compilation" not in self.supported_extra_tags:
            return None
        if value:
            self.set_tag("compilation", value)
            self.mark_as_modified("compilation")
        elif self.has_tag("compilation"):
            self.delete_tag("compilation")

    @GObject.Property(type=str)
    def composer(self):
        if "composer" in self.supported_extra_tags:
            return self.get_tag("composer")
        return None

    @composer.setter
    def composer(self, value):
        if "composer" not in self.supported_extra_tags:
            return None
        if value:
            self.set_tag("composer", value)
            self.mark_as_modified("composer")
        elif self.has_tag("composer"):
            self.delete_tag("composer")

    @GObject.Property(type=str)
    def copyright(self):
        if "copyright" in self.supported_extra_tags:
            return self.get_tag("copyright")
        return None

    @copyright.setter
    def copyright(self, value):
        if "copyright" not in self.supported_extra_tags:
            return None
        if value:
            self.set_tag("copyright", value)
            self.mark_as_modified("copyright")
        elif self.has_tag("copyright"):
            self.delete_tag("copyright")

    @GObject.Property(type=str)
    def encodedby(self):
        if "encodedby" in self.supported_extra_tags:
            return self.get_tag("encodedby")
        return None

    @encodedby.setter
    def encodedby(self, value):
        if "encodedby" not in self.supported_extra_tags:
            return None
        if value:
            self.set_tag("encodedby", value)
            self.mark_as_modified("encodedby")
        elif self.has_tag("encodedby"):
            self.delete_tag("encodedby")

    @GObject.Property(type=str)
    def mood(self):
        if "mood" in self.supported_extra_tags:
            return self.get_tag("mood")
        return None

    @mood.setter
    def mood(self, value):
        if "mood" not in self.supported_extra_tags:
            return None
        if value:
            self.set_tag("mood", value)
            self.mark_as_modified("mood")
        elif self.has_tag("mood"):
            self.delete_tag("mood")

    @GObject.Property(type=str)
    def conductor(self):
        if "conductor" in self.supported_extra_tags:
            return self.get_tag("conductor")
        return None

    @conductor.setter
    def conductor(self, value):
        if "conductor" not in self.supported_extra_tags:
            return None
        if value:
            self.set_tag("conductor", value)
            self.mark_as_modified("conductor")
        elif self.has_tag("conductor"):
            self.delete_tag("conductor")

    @GObject.Property(type=str)
    def arranger(self):
        if "arranger" in self.supported_extra_tags:
            return self.get_tag("arranger")
        return None

    @arranger.setter
    def arranger(self, value):
        if "arranger" not in self.supported_extra_tags:
            return None
        if value:
            self.set_tag("arranger", value)
            self.mark_as_modified("arranger")
        elif self.has_tag("arranger"):
            self.delete_tag("arranger")

    @GObject.Property(type=int)
    def discnumber(self):
        if "discnumber" in self.supported_extra_tags:
            value = self.get_tag("discnumber")
            if value:
                return safe_int(value)
        return None

    @discnumber.setter
    def discnumber(self, value):
        if "discnumber" not in self.supported_extra_tags:
            return None
        if value:
            self.set_tag("discnumber", safe_int(value))
            self.mark_as_modified("discnumber")
        elif self.has_tag("discnumber"):
            self.delete_tag("discnumber")

    @GObject.Property(type=str)
    def publisher(self):
        if "publisher" in self.supported_extra_tags:
            return self.get_tag("publisher")
        return None

    @publisher.setter
    def publisher(self, value):
        if "publisher" not in self.supported_extra_tags:
            return None
        if value:
            self.set_tag("publisher", value)
            self.mark_as_modified("publisher")
        elif self.has_tag("publisher"):
            self.delete_tag("publisher")

    @GObject.Property(type=str)
    def isrc(self):
        if "isrc" in self.supported_extra_tags:
            return self.get_tag("isrc")
        return None

    @isrc.setter
    def isrc(self, value):
        if "isrc" not in self.supported_extra_tags:
            return None
        if value:
            self.set_tag("isrc", value)
            self.mark_as_modified("isrc")
        elif self.has_tag("isrc"):
            self.delete_tag("isrc")

    @GObject.Property(type=str)
    def language(self):
        if "language" in self.supported_extra_tags:
            return self.get_tag("language")
        return None

    @language.setter
    def language(self, value):
        if "language" not in self.supported_extra_tags:
            return None
        if value:
            self.set_tag("language", value)
            self.mark_as_modified("language")
        elif self.has_tag("language"):
            self.delete_tag("language")

    @GObject.Property(type=str)
    def discsubtitle(self):
        if "discsubtitle" in self.supported_extra_tags:
            return self.get_tag("discsubtitle")
        return None

    @discsubtitle.setter
    def discsubtitle(self, value):
        if "discsubtitle" not in self.supported_extra_tags:
            return None
        if value:
            self.set_tag("discsubtitle", value)
            self.mark_as_modified("discsubtitle")
        elif self.has_tag("discsubtitle"):
            self.delete_tag("discsubtitle")

    # Sort order tags

    @GObject.Property(type=str)
    def albumartistsort(self):
        if "albumartistsort" in self.supported_extra_tags:
            return self.get_tag("albumartistsort")
        return None

    @albumartistsort.setter
    def albumartistsort(self, value):
        if "albumartistsort" not in self.supported_extra_tags:
            return None
        if value:
            self.set_tag("albumartistsort", value)
            self.mark_as_modified("albumartistsort")
        elif self.has_tag("albumartistsort"):
            self.delete_tag("albumartistsort")

    @GObject.Property(type=str)
    def albumsort(self):
        if "albumsort" in self.supported_extra_tags:
            return self.get_tag("albumsort")
        return None

    @albumsort.setter
    def albumsort(self, value):
        if "albumsort" not in self.supported_extra_tags:
            return None
        if value:
            self.set_tag("albumsort", value)
            self.mark_as_modified("albumsort")
        elif self.has_tag("albumsort"):
            self.delete_tag("albumsort")

    @GObject.Property(type=str)
    def composersort(self):
        if "composersort" in self.supported_extra_tags:
            return self.get_tag("composersort")
        return None

    @composersort.setter
    def composersort(self, value):
        if "composersort" not in self.supported_extra_tags:
            return None
        if value:
            self.set_tag("composersort", value)
            self.mark_as_modified("composersort")
        elif self.has_tag("composersort"):
            self.delete_tag("composersort")

    @GObject.Property(type=str)
    def artistsort(self):
        if "artistsort" in self.supported_extra_tags:
            return self.get_tag("artistsort")
        return None

    @artistsort.setter
    def artistsort(self, value):
        if "artistsort" not in self.supported_extra_tags:
            return None
        if value:
            self.set_tag("artistsort", value)
            self.mark_as_modified("artistsort")
        elif self.has_tag("artistsort"):
            self.delete_tag("artistsort")

    @GObject.Property(type=str)
    def titlesort(self):
        if "titlesort" in self.supported_extra_tags:
            return self.get_tag("titlesort")
        return None

    @titlesort.setter
    def titlesort(self, value):
        if "titlesort" not in self.supported_extra_tags:
            return None
        if value:
            self.set_tag("titlesort", value)
            self.mark_as_modified("titlesort")
        elif self.has_tag("titlesort"):
            self.delete_tag("titlesort")

    @GObject.Property(type=str)
    def url(self):
        if "url" in self.supported_extra_tags:
            return self.get_tag("url")
        return None

    @url.setter
    def url(self, value):
        if "url" not in self.supported_extra_tags:
            return None
        if value:
            self.set_tag("url", value)
            self.mark_as_modified("url")
        elif self.has_tag("url"):
            self.delete_tag("url")

    # MusicBrainz tags

    @GObject.Property(type=str)
    def musicbrainz_artistid(self):
        if "musicbrainz_artistid" in self.supported_extra_tags:
            return self.get_tag("musicbrainz_artistid")
        return None

    @musicbrainz_artistid.setter
    def musicbrainz_artistid(self, value):
        if "musicbrainz_artistid" not in self.supported_extra_tags:
            return None
        if value:
            self.set_tag("musicbrainz_artistid", value)
            self.mark_as_modified("musicbrainz_artistid")
        elif self.has_tag("musicbrainz_artistid"):
            self.delete_tag("musicbrainz_artistid")

    @GObject.Property(type=str)
    def musicbrainz_albumid(self):
        if "musicbrainz_albumid" in self.supported_extra_tags:
            return self.get_tag("musicbrainz_albumid")
        return None

    @musicbrainz_albumid.setter
    def musicbrainz_albumid(self, value):
        if "musicbrainz_albumid" not in self.supported_extra_tags:
            return None
        if value:
            self.set_tag("musicbrainz_albumid", value)
            self.mark_as_modified("musicbrainz_albumid")
        elif self.has_tag("musicbrainz_albumid"):
            self.delete_tag("musicbrainz_albumid")

    @GObject.Property(type=str)
    def musicbrainz_albumartistid(self):
        if "musicbrainz_albumartistid" in self.supported_extra_tags:
            return self.get_tag("musicbrainz_albumartistid")
        return None

    @musicbrainz_albumartistid.setter
    def musicbrainz_albumartistid(self, value):
        if "musicbrainz_albumartistid" not in self.supported_extra_tags:
            return None
        if value:
            self.set_tag("musicbrainz_albumartistid", value)
            self.mark_as_modified("musicbrainz_albumartistid")
        elif self.has_tag("musicbrainz_albumartistid"):
            self.delete_tag("musicbrainz_albumartistid")

    @GObject.Property(type=str)
    def musicbrainz_trackid(self):
        if "musicbrainz_trackid" in self.supported_extra_tags:
            return self.get_tag("musicbrainz_trackid")
        return None

    @musicbrainz_trackid.setter
    def musicbrainz_trackid(self, value):
        if "musicbrainz_trackid" not in self.supported_extra_tags:
            return None
        if value:
            self.set_tag("musicbrainz_trackid", value)
            self.mark_as_modified("musicbrainz_trackid")
        elif self.has_tag("musicbrainz_trackid"):
            self.delete_tag("musicbrainz_trackid")

    @GObject.Property(type=str)
    def musicbrainz_recordingid(self):
        if "musicbrainz_recordingid" in self.supported_extra_tags:
            return self.get_tag("musicbrainz_recordingid")
        return None

    @musicbrainz_recordingid.setter
    def musicbrainz_recordingid(self, value):
        if "musicbrainz_recordingid" not in self.supported_extra_tags:
            return None
        if value:
            self.set_tag("musicbrainz_recordingid", value)
            self.mark_as_modified("musicbrainz_recordingid")
        elif self.has_tag("musicbrainz_recordingid"):
            self.delete_tag("musicbrainz_recordingid")

    @GObject.Property(type=str)
    def musicbrainz_releasegroupid(self):
        if "musicbrainz_releasegroupid" in self.supported_extra_tags:
            return self.get_tag("musicbrainz_releasegroupid")
        return None

    @musicbrainz_releasegroupid.setter
    def musicbrainz_releasegroupid(self, value):
        if "musicbrainz_releasegroupid" not in self.supported_extra_tags:
            return None
        if value:
            self.set_tag("musicbrainz_releasegroupid", value)
            self.mark_as_modified("musicbrainz_releasegroupid")
        elif self.has_tag("musicbrainz_releasegroupid"):
            self.delete_tag("musicbrainz_releasegroupid")

    @GObject.Property(type=str)
    def none(self):
        return ""

    @none.setter
    def none(self, value):
        return ""



# SPDX-License-Identifier: MIT
# (c) 2023 knuxify and Ear Tag contributors

from gi.repository import GObject, Gtk, GLib, Gdk
import os.path

from . import APP_GRESOURCE_PATH


@Gtk.Template(resource_path=f"{APP_GRESOURCE_PATH}/ui/filelistitem.ui")
class EartagFileListItem(Gtk.Box):
    __gtype_name__ = "EartagFileListItem"

    status_icon_stack = Gtk.Template.Child()
    modified_icon = Gtk.Template.Child()
    error_icon = Gtk.Template.Child()

    coverart_image = Gtk.Template.Child()
    title_label = Gtk.Template.Child()
    filename_label = Gtk.Template.Child()

    cover_edit_stack = Gtk.Template.Child()
    select_button = Gtk.Template.Child()
    remove_button = Gtk.Template.Child()
    suffixes = Gtk.Template.Child()

    def __init__(self, filelist):
        super().__init__()
        self.file = None
        self._title = None
        self._selected = None
        self.filelist = filelist

        self.bindings = []
        self.connections = []

        if self.filelist.selection_mode:
            self.show_selection_button()

        self.file_manager = filelist.file_manager
        self._fm_conn = self.file_manager.connect("selection-changed", self.update_selected_status)
        self._fl_conn = self.filelist.connect("notify::selection-mode", self.toggle_selection_mode)

        self._selected_bind = self.bind_property(
            "selected",
            self.select_button,
            "active",
            GObject.BindingFlags.BIDIRECTIONAL | GObject.BindingFlags.SYNC_CREATE,
        )

        self.long_press_gesture = Gtk.GestureLongPress.new()
        self._long_press_connect = self.long_press_gesture.connect("pressed", self.on_long_press)
        self.add_controller(self.long_press_gesture)

    def bind_to_file(self, file):
        if self.file:
            self.unbind_from_file()
        self.file = file

        # self.bindings.append(
        # 	self.file.bind_property(
        # 		"title", self, "title", GObject.BindingFlags.SYNC_CREATE
        # 	)
        # )
        self.connections.append(self.file.connect("notify::title", self.on_title_notify))
        self.on_title_notify()

        self.bindings.append(
            self.file.bind_property("path", self, "filename", GObject.BindingFlags.SYNC_CREATE)
        )

        self.bindings.append(
            self.file.bind_property(
                "is-modified",
                self.modified_icon,
                "visible",
                GObject.BindingFlags.SYNC_CREATE,
            )
        )
        self.bindings.append(
            self.file.bind_property(
                "has-error",
                self.error_icon,
                "visible",
                GObject.BindingFlags.SYNC_CREATE,
            )
        )
        self.connections.append(self.file.connect("notify::has-error", self.handle_error))
        self.filename_label.set_label(os.path.basename(file.path))
        self.coverart_image.bind_to_file(file)

        self.update_selected_status()

    def unbind_from_file(self, file=None):
        if not self.file or (file and file != self.file):
            return

        for b in self.bindings:
            b.unbind()
        self.bindings = []

        for conn in self.connections:
            self.file.disconnect(conn)
        self.connections = []

        self.coverart_image.unbind_from_file(self.file)

        self.file = None

    def teardown(self, *args):
        self.unbind_from_file()

        self.filelist.disconnect(self._fl_conn)
        del self._fl_conn
        del self.fileview

        self.file_manager.disconnect(self._fm_conn)
        del self._fm_conn
        del self.file_manager

        self.disconnect(self._selected_bind)
        del self._selected_bind

        self.long_press_gesture.disconnect(self._long_press_connect)
        self.remove_controller(self.long_press_gesture)
        del self.long_press_gesture

    def on_title_notify(self, *args):
        # Because calling notify on a property doesn't update bind_property,
        # which we need to do for delete_tag on the file, we have to connect
        # to the notify signal instead.
        self.props.title = self.file.props.title

    def handle_error(self, *args):
        if self.file.has_error:
            self.status_icon_stack.set_visible_child(self.error_icon)
        else:
            self.status_icon_stack.set_visible_child(self.modified_icon)

    def add_suffix(self, widget):
        """Adds a suffix widget."""
        self.suffixes.append(widget)

    @GObject.Property(type=bool, default=False)
    def selected(self):
        if not self.file:
            return False
        return self._selected

    @selected.setter
    def selected(self, value):
        if not self.file:
            return
        self._selected = value
        if value and not self.file_manager.is_selected(self.file):
            self.file_manager.select_file(self.file)
        elif not value and self.file_manager.is_selected(self.file):
            self.file_manager.unselect_file(self.file)

    def update_selected_status(self, *args):
        if not self.file:
            return
        self._selected = self.file_manager.is_selected(self.file)
        self.notify("selected")

    def on_long_press(self, *args):
        if self.filelist.props.selection_mode is False:
            self.filelist._ignore_activate = True
            if not self.props.selected:
                self.file_manager.select_file(self.file, unselect_other=True)
                self._selected = True
                self.notify("selected")
            self.filelist.force_switch_into_selection_mode()

    @Gtk.Template.Callback()
    def remove_item(self, *args):
        # HACK: When the remove button is pressed, the long-press gesture gets
        # triggered. (It seems that the click is not freed correctly?)
        # Might be the same issue as https://gitlab.gnome.org/GNOME/gtk/-/issues/5794
        self.props.sensitive = False
        self.props.sensitive = True
        if self.file_manager.remove_files([self.file]):
            self.on_destroy()

    @GObject.Property(type=str)
    def title(self):
        return self._title

    @title.setter
    def title(self, value):
        # TRANSLATORS: Placeholder for file sidebar items with no title set
        self.title_label.set_label(value or _("(No title)"))

    @GObject.Property(type=str)
    def filename(self):
        return self._filename

    @filename.setter
    def filename(self, value):
        self.filename_label.set_label(os.path.basename(value))
        self._filename = os.path.basename(value)

    def show_selection_button(self, *args):
        self.cover_edit_stack.set_visible_child(self.select_button)

    def hide_selection_button(self, *args):
        self.cover_edit_stack.set_visible_child(self.coverart_image)

    def toggle_selection_mode(self, *args):
        if self.filelist.selection_mode:
            self.show_selection_button()
        else:
            self.hide_selection_button()


class EartagFileList(Gtk.ListView):
    """List of opened tracks."""

    __gtype_name__ = "EartagFileList"

    def __init__(self):
        super().__init__()
        self._selection_mode = False
        self._ignore_unselect = False
        self.file_manager = None

        # See on_activate function for explaination
        self._ignore_activate = False
        self.connect("activate", self.on_activate)
        self.key_controller = Gtk.EventControllerKey.new()
        self.key_controller.connect("key-pressed", self.key_pressed)
        self.key_controller.connect("key-released", self.key_released)
        self.add_controller(self.key_controller)
        self.shift_state = False

        self.sidebar_factory = Gtk.SignalListItemFactory()
        self.sidebar_factory.connect("setup", self.setup)
        self.sidebar_factory.connect("bind", self.bind)
        self.sidebar_factory.connect("unbind", self.unbind)
        self.sidebar_factory.connect("teardown", self.teardown)
        self.set_factory(self.sidebar_factory)

    def set_file_manager(self, file_manager):
        self.file_manager = file_manager
        self.set_model(self.file_manager.selected_files)
        self.file_manager.connect("selection-changed", self.switch_into_selection_mode)

    def setup(self, factory, list_item):
        list_item.set_child(EartagFileListItem(self))

    def bind(self, factory, list_item):
        child = list_item.get_child()
        file = list_item.get_item()
        child.bind_to_file(file)

    def unbind(self, factory, list_item):
        child = list_item.get_child()
        if child:
            child.unbind_from_file()

    def teardown(self, factory, list_item):
        child = list_item.get_child()
        if child:
            child.teardown()

    @GObject.Property(type=bool, default=False)
    def selection_mode(self):
        """Whether the sidebar is in selection mode or not."""
        return self._selection_mode

    @selection_mode.setter
    def selection_mode(self, value):
        self._selection_mode = value
        if not self.file_manager:
            return
        if value is True:
            self.action_set_enabled("list.select-item", False)
            self.props.single_click_activate = True
        else:
            self.action_set_enabled("list.select-item", True)
            self.props.single_click_activate = False
            selection = self.file_manager.selected_files.get_selection()
            if selection.get_size() > 1:
                self.file_manager.selected_files.select_item(selection.get_nth(0), True)

    def toggle_selection_mode(self):
        self.props.selection_mode = not self.props.selection_mode

    def switch_into_selection_mode(self, *args):
        """
        If multiple files have been selected (by holding down Ctrl or Shift) and
        we're not in multiple selection mode, toggle it on.
        """
        if self.file_manager.get_n_selected() >= 2 and not self.props.selection_mode:
            self.props.selection_mode = True

    def force_switch_into_selection_mode(self, *args):
        """
        Force-enable selection mode.
        """
        if not self.props.selection_mode:
            self.props.selection_mode = True

    def remove_selected(self):
        """Removes selected files."""
        if not self.file_manager:
            return
        old_selected = self.file_manager.selected_files_list.copy()
        self.file_manager.remove_files(old_selected)

    def on_activate(self, list, index):
        if self._ignore_activate is True:
            self._ignore_activate = False
            return

        if self.props.selection_mode:
            # Gtk.MultiSelection doesn't have a way to force every click
            # to count a selection - instead, a single click unselects
            # all other items. This isn't the behavior we want, so instead:
            # - We switch into single_click_activate (which makes selections
            # happen on hover, and activations on a single click, instead
            # of selections happening on a single-click and activations on
            # double-click)
            # - Then, we disable the list.select-item action to prevent the
            # items from being selected on hover. Instead:
            # - We re-implement the default selection behavior manually,
            # here in on_activate. We still call the underlying action so
            # that we don't have to re-write all the code for handling
            # "extend" selections (when Shift is pressed).

            extend = self.shift_state

            self.action_set_enabled("list.select-item", True)
            self.activate_action(
                "list.select-item", GLib.Variant("(ubb)", (index, not extend, extend))
            )
            self.action_set_enabled("list.select-item", False)

    def key_pressed(self, controller, keyval, keycode, state):
        # Used for selection mode, see on_activate.
        if state & Gdk.ModifierType.SHIFT_MASK or keyval in (
            Gdk.KEY_Shift_L,
            Gdk.KEY_Shift_R,
        ):
            self.shift_state = True

        if keyval == Gdk.KEY_Delete:
            self.remove_selected()

    def key_released(self, controller, keyval, keycode, state):
        # Used for selection mode, see on_activate.
        if state & Gdk.ModifierType.SHIFT_MASK or keyval in (
            Gdk.KEY_Shift_L,
            Gdk.KEY_Shift_R,
        ):
            self.shift_state = False



# SPDX-License-Identifier: MIT
# (c) 2023 knuxify and Ear Tag contributors

from gi.repository import Adw, Gtk, Gio, GObject, GdkPixbuf

import asyncio
import os
import html
import traceback
import re

from .musicbrainz import (
    acoustid_identify_file,
    MusicBrainzRecording,
    MusicBrainzRelease,
)
from .logger import logger
from .utils import find_in_model
from .utils.asynctask import EartagAsyncTask
from .utils.extracttags import extract_tags_from_filename
from .utils.widgets import EartagModelExpanderRow
from .backends.file import EartagFile
from . import APP_GRESOURCE_PATH


MBID_REGEX = r"[0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{12}"
RECORDING_URL_REGEX = r"^(https?://)?musicbrainz.org/recording/(?P<mbid>" + MBID_REGEX + ")/?$"


@Gtk.Template(resource_path=f"{APP_GRESOURCE_PATH}/ui/identify/coverimage.ui")
class EartagIdentifyCoverImage(Gtk.Stack):
    __gtype_name__ = "EartagIdentifyCoverImage"

    no_cover = Gtk.Template.Child()
    cover_image = Gtk.Template.Child()
    loading_icon = Gtk.Template.Child()

    loading = GObject.Property(type=bool, default=False)
    cover_path = GObject.Property(type=str, default="")

    def __init__(self):
        super().__init__()
        self._cover_path = None
        self.connect("notify::loading", self.update)
        self.connect("notify::cover-path", self.update)

    def update(self, *args):
        """Update the cover according to the current state."""
        if self.props.loading:
            self.set_visible_child(self.loading_icon)
            return

        path = self.props.cover_path

        if not path or not os.path.exists(path):
            self.set_visible_child(self.no_cover)
            return

        pixbuf = GdkPixbuf.Pixbuf.new_from_file_at_scale(path, 48, 48, True)

        self.cover_image.set_from_pixbuf(pixbuf)
        self.set_visible_child(self.cover_image)


class EartagIdentifyReleaseRow(EartagModelExpanderRow):
    """
    Representation of MusicBrainz releases for the ItemRow
    dropdowns.
    """

    __gtype_name__ = "EartagIdentifyReleaseRow"

    def __init__(self, parent, release=None):
        super().__init__()
        self._bindings = []
        self._connections = []
        self.parent = parent
        self.release = None
        self.obj = None

        # Unfortunately we can't set the parent to EartagModelExpanderRow
        # in a template for some weird reason, so we have to set this widget
        # up manually here:

        self.add_css_class("identify-release-row")

        self.cover_image = EartagIdentifyCoverImage()
        self.cover_image.set_hexpand(False)
        self.add_prefix(self.cover_image)

        self.apply_checkbox = Gtk.CheckButton()
        self.apply_checkbox.set_active(True)
        self.apply_checkbox.set_sensitive(False)
        self.apply_checkbox.set_valign(Gtk.Align.CENTER)
        self.apply_checkbox.connect("notify::active", self.toggle_row_checkboxes)
        self.apply_checkbox.add_css_class("selection-mode")
        self.apply_checkbox.set_tooltip_text(_("Apply identified data"))
        self.add_suffix(self.apply_checkbox)

        if release:
            self.bind_to_release(release)

        self._rec_filter = Gtk.CustomFilter()
        self._rec_filter.set_filter_func(self._rec_filter_func)
        self._rec_filter_model = Gtk.FilterListModel(
            model=self.parent.recordings_model, filter=self._rec_filter
        )

        self._rec_sorter = Gtk.CustomSorter()
        self._rec_sorter.set_sort_func(self._rec_sorter_func)
        self._rec_sorter_model = Gtk.SortListModel(
            model=self._rec_filter_model, sorter=self._rec_sorter
        )

        self.bind_model(self._rec_sorter_model, self.row_create)

        # Release chooser filter setup
        self._rel_model = Gio.ListStore(item_type=MusicBrainzRelease)

        self.release_popover = Gtk.Popover()
        self.release_popover_scrolled_window = Gtk.ScrolledWindow()
        self.release_popover_list = Gtk.ListBox()
        self.release_popover_list.bind_model(self._rel_model, self.rel_row_create)
        self.release_popover_list.add_css_class("boxed-list")
        self.release_popover_list.set_selection_mode(Gtk.SelectionMode.NONE)
        self.release_popover_scrolled_window.set_policy(
            Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC
        )
        self.release_popover_scrolled_window.set_propagate_natural_height(True)
        self.release_popover_scrolled_window.set_propagate_natural_width(True)
        self.release_popover_scrolled_window.set_child(self.release_popover_list)
        self.release_popover.set_child(self.release_popover_scrolled_window)
        self._relswitch_first_row = None

        self.release_url_button = Gtk.Button()
        self.release_url_button.set_icon_name("external-link-symbolic")
        # TRANSLATORS: Tooltip for button to open release info in MusicBrainz identification dialog.
        self.release_url_button.set_tooltip_text(_("See release on MusicBrainz"))
        self.release_url_button.set_valign(Gtk.Align.CENTER)
        self.release_url_button.connect("clicked", self.open_release_url)
        self.release_url_button.add_css_class("flat")
        self.release_url_launcher = Gtk.UriLauncher.new("")
        self.add_suffix(self.release_url_button)

        self.release_popover_toggle = Gtk.MenuButton(popover=self.release_popover)
        self.release_popover_toggle.set_valign(Gtk.Align.CENTER)
        self.release_popover_toggle.set_visible(False)
        # TRANSLATORS: Tooltip for release switcher button in MusicBrainz identification dialog.
        # This allows the user to switch between different releases of an album, EP, etc.
        self.release_popover_toggle.set_tooltip_text(_("Other releases"))
        self.release_popover_toggle.set_icon_name("view-more-symbolic")
        self.release_popover_toggle.add_css_class("flat")
        self.release_popover_toggle.connect("notify::active", self.download_alt_release_thumbnails)
        self.add_suffix(self.release_popover_toggle)

    def bind_to_release(self, release):
        """Takes a MusicBrainzRelease and binds to it."""
        if self.release:
            self.unbind()

        self.release = release

        self._bindings = [
            self.release.bind_property(
                "thumbnail_path",
                self.cover_image,
                "cover_path",
                GObject.BindingFlags.SYNC_CREATE,
            ),
            self.release.bind_property(
                "thumbnail_loaded",
                self.cover_image,
                "loading",
                GObject.BindingFlags.SYNC_CREATE | GObject.BindingFlags.INVERT_BOOLEAN,
            ),
        ]
        self._connections = [
            self.release.connect("notify::title", self.update_title),
            self.release.connect("notify::disambiguation", self.update_title),
            self.release.connect("notify::artist", self.update_subtitle),
            self.release.connect("notify::releasedate", self.update_subtitle),
        ]
        self.update_title()
        self.update_subtitle()

        release.queue_thumbnail_download()

    def unbind(self):
        for binding in self._bindings:
            binding.unbind()

        for conn in self._connections:
            if conn:
                self.release.disconnect(conn)

        self.release = None

    def update_title(self, *args):
        title = html.escape(self.release.title)
        if self.release.disambiguation:
            opacity = 0.55
            if Adw.StyleManager.get_default().get_high_contrast():
                opacity = 0.9
            title += (
                f' <span alpha="{int(opacity * 100)}%">('
                + html.escape(self.release.disambiguation)
                + ")</span>"
            )
        self.set_title(title)

    def update_subtitle(self, *args):
        self._subtitle = html.escape(self.release.artist)
        if self.release.releasedate:
            self._subtitle += " • " + self.release.releasedate
        self.set_subtitle(self._subtitle)

    def update_filter(self):
        self._filter_changed = False
        self._rec_filter.changed(Gtk.FilterChange.DIFFERENT)
        self._filter_changed = True

    def _rec_filter_func(self, recording):
        try:
            assert recording.release
        except (AttributeError, ValueError, AssertionError):
            return False
        return recording.release.release_id == self.release.release_id

    def _rec_sorter_func(self, rec1, rec2, *args):
        return rec1.tracknumber - rec2.tracknumber

    def row_create(self, recording, *args):
        row = EartagIdentifyRecordingRow(self, recording)
        row.apply_checkbox.connect("notify::active", self.toggle_make_inconsistent)
        return row

    def toggle_apply_sensitivity(self, value):
        n = 0
        row = self.get_row_at_index(n)
        while row:
            row.apply_checkbox.set_sensitive(value)
            n += 1
            row = self.get_row_at_index(n)
        self.apply_checkbox.set_sensitive(value)
        self.release_popover_toggle.set_sensitive(value)
        self.release_popover_toggle.set_visible(bool(self._rel_model.get_n_items() > 1))

    def toggle_row_checkboxes(self, toggle, *args):
        toggle.set_inconsistent(False)
        n = 0
        row = self.get_row_at_index(n)
        while row:
            row.apply_checkbox.set_active(toggle.props.active)
            n += 1
            row = self.get_row_at_index(n)

    def toggle_make_inconsistent(self, *args):
        n = 0
        row = self.get_row_at_index(n)
        active = row.apply_checkbox.get_active()
        while row:
            if row.apply_checkbox.get_active() != active:
                self.apply_checkbox.set_inconsistent(True)
                return
            n += 1
            row = self.get_row_at_index(n)
        self.apply_checkbox.set_inconsistent(False)
        if active != self.apply_checkbox.props.active:
            self.apply_checkbox.set_active(active)

    def refresh_alternative_releases(self, releases):
        """
        Used by the identification process to fill in the release switcher
        button model.
        """
        self.release_popover_toggle.set_visible(bool(releases))
        self._rel_model.splice(0, self._rel_model.get_n_items(), releases)

    def rel_row_create(self, rel, *args):
        row = EartagIdentifyAltReleaseRow(self, rel)
        if self._relswitch_first_row:
            row.apply_checkbox.set_active(False)
            row.apply_checkbox.set_group(self._relswitch_first_row.apply_checkbox)
            self._relswitch_first_row.set_sensitive(True)
        else:
            self._relswitch_first_row = row
            row.set_sensitive(False)
        if rel == self.release:
            row.apply_checkbox.set_active(True)
        row.apply_checkbox.connect("notify::active", self.set_release_from_selector, rel.release_id)
        return row

    def download_alt_release_thumbnails(self, *args):
        """
        Start downloading alternative release thumbnails when the popover
        is opened.
        """
        if not self.release_popover_toggle.props.active:
            return

        for rel in self._rel_model:
            rel.queue_thumbnail_download()

    def set_release_from_selector(self, _t1, _t2, rel_id):
        rel = None
        for rel in self._rel_model:
            if rel.release_id == rel_id:
                break

        if self.release:
            del self.parent.release_rows[self.release.release_id]

        for rec in self._rec_sorter_model:
            for _rel in rec.available_releases:
                if _rel.release_id == rel.release_id:
                    rec.release = _rel
                    break

        if rel_id in self.parent.release_rows:
            self.parent.content_listbox.remove(self)
            self.parent.release_rows[rel_id].update_filter()
            self.parent.release_rows[rel_id].toggle_apply_sensitivity(True)
        else:
            self.bind_to_release(rel)
            self.parent.release_rows[rel_id] = self
            self.update_filter()
            self.toggle_apply_sensitivity(True)

    def open_release_url(self, *args):
        """Open the currently bound release's MusicBrainz page."""

        async def _open(self):
            self.release_url_launcher.props.uri = (
                f"https://musicbrainz.org/release/{self.release.release_id}"
            )
            await self.release_url_launcher.launch(self.get_root())

        asyncio.create_task(_open(self))


@Gtk.Template(resource_path=f"{APP_GRESOURCE_PATH}/ui/identify/altreleaserow.ui")
class EartagIdentifyAltReleaseRow(Adw.ActionRow):
    """
    Representation of releases for the release switcher dropdown.
    """

    __gtype_name__ = "EartagIdentifyAltReleaseRow"

    cover_image = Gtk.Template.Child()
    apply_checkbox = Gtk.Template.Child()

    def __init__(self, parent, release):
        super().__init__()
        self._bindings = []
        self._connections = []
        self.release = release

        self.bind_to_release(release)

    def bind_to_release(self, release):
        self.release = release

        self._bindings = [
            self.release.bind_property(
                "thumbnail_path",
                self.cover_image,
                "cover_path",
                GObject.BindingFlags.SYNC_CREATE,
            ),
            self.release.bind_property(
                "thumbnail_loaded",
                self.cover_image,
                "loading",
                GObject.BindingFlags.SYNC_CREATE | GObject.BindingFlags.INVERT_BOOLEAN,
            ),
        ]
        self._connections = [
            self.release.connect("notify::title", self.update_title),
            self.release.connect("notify::artist", self.update_subtitle),
        ]
        self.update_title()
        self.update_subtitle()

    def unbind(self, *args):
        if self.release:
            for binding in self._bindings:
                binding.unbind()

            for conn in self._connections:
                self.release.disconnect(conn)

        self._connections = []
        self._bindings = []

        self.release = None
        self.parent = None

    def update_title(self, *args):
        title = html.escape(self.release.title)
        if self.release.disambiguation:
            opacity = 0.55
            if Adw.StyleManager.get_default().get_high_contrast():
                opacity = 0.9
            title += (
                f' <span alpha="{int(opacity * 100)}%">('
                + html.escape(self.release.disambiguation)
                + ")</span>"
            )
        self.set_title(title)

    def update_subtitle(self, *args):
        self._subtitle = html.escape(self.release.artist)
        if self.release.releasedate:
            self._subtitle += " • " + self.release.releasedate
        self.set_subtitle(self._subtitle)


@Gtk.Template(resource_path=f"{APP_GRESOURCE_PATH}/ui/identify/filerow.ui")
class EartagIdentifyFileRow(Adw.ActionRow):
    """
    Representation of files for the identify dialog.
    """

    __gtype_name__ = "EartagIdentifyFileRow"

    cover_image = Gtk.Template.Child()

    suffix_stack = Gtk.Template.Child()
    loading_icon = Gtk.Template.Child()
    not_found_icon = Gtk.Template.Child()

    recording_override_popover = Gtk.Template.Child()
    recording_override_entry = Gtk.Template.Child()
    recording_override_apply_button = Gtk.Template.Child()

    def __init__(self, file, parent):
        super().__init__()
        self._connections = []
        self.file = None
        self.parent = parent

        self.connect("destroy", self.unbind)

        self.bind_to_file(file)

        self.validate_recording_override_entry()

    def bind_to_file(self, file):
        if self.file:
            self.unbind()

        self.file = file
        self.cover_image.bind_to_file(file)

        self._connections = [
            self.file.connect("notify::title", self.update_title),
            self.file.connect("notify::artist", self.update_subtitle),
            self.file.connect("notify::album", self.update_subtitle),
        ]
        self.update_title()
        self.update_subtitle()

    def unbind(self, *args):
        for conn in self._connections:
            self.file.disconnect(conn)
        self.cover_image.unbind_from_file()

        self.file = None

    def update_title(self, *args):
        self.set_title(html.escape(self.file.title))

    def update_subtitle(self, *args):
        self._subtitle = (
            f"{self.file.artist or 'N/A'} • {self.file.album or 'N/A'}"
            + f" ({os.path.basename(self.file.path)})"
        )
        self.set_subtitle(html.escape(self._subtitle))

    def start_loading(self):
        self.suffix_stack.set_visible(True)
        self.suffix_stack.set_visible_child(self.loading_icon)

    def mark_as_unidentified(self):
        self.suffix_stack.set_visible_child(self.not_found_icon)

    def get_mbid_from_recording_override_entry(self):
        # Reuse the IdentifyRecordingRow function for convenience
        return EartagIdentifyRecordingRow.get_mbid_from_recording_override_entry(self)

    @Gtk.Template.Callback()
    def validate_recording_override_entry(self, *args):
        # Reuse the IdentifyRecordingRow function for convenience
        return EartagIdentifyRecordingRow.validate_recording_override_entry(self)

    @Gtk.Template.Callback()
    def set_recording_override_from_entry(self, *args):
        """Apply the recording ID or URL from the recording override entry."""

        async def _set_rec_override(self):
            mbid = self.get_mbid_from_recording_override_entry()
            if not mbid:
                return

            self.recording_override_popover.props.sensitive = False

            rec = await MusicBrainzRecording.new_for_id(mbid)

            self.parent._identify_set_recording(self.file.id, rec)
            self.parent.on_identify_done(None, show_toast=False)

            self.recording_override_popover.props.sensitive = True

        asyncio.create_task(_set_rec_override(self))


@Gtk.Template(resource_path=f"{APP_GRESOURCE_PATH}/ui/identify/recordingrow.ui")
class EartagIdentifyRecordingRow(Adw.ActionRow):
    """
    Representation of recordings for the identify dialog.
    """

    __gtype_name__ = "EartagIdentifyRecordingRow"

    apply_checkbox = Gtk.Template.Child()
    recording_url_button = Gtk.Template.Child()

    recording_override_popover = Gtk.Template.Child()
    recording_override_entry = Gtk.Template.Child()
    recording_override_apply_button = Gtk.Template.Child()

    def __init__(self, parent, recording):
        super().__init__()
        self._bindings = []
        self._connections = []
        self.recording = None
        file_id = None
        self.parent = parent
        assert self.parent
        for _file_id, rec in self.parent.parent.recordings.items():
            if rec.recording_id == recording.recording_id:
                file_id = _file_id
                break
        self.file_id = file_id

        self.file_name = ""
        if file_id:
            for file in self.parent.parent.files:
                if file.id == self.file_id:
                    self.file_name = os.path.basename(file.path)
                    break

        self.recording_url_launcher = Gtk.UriLauncher.new("")

        self.connect("destroy", self.unbind)

        self.bind_to_recording(recording)

        self.validate_recording_override_entry()

    def bind_to_recording(self, recording):
        self.recording = recording

        self._connections = [
            self.recording.connect("notify::title", self.update_title),
            self.recording.connect("notify::artist", self.update_subtitle),
            self.recording.connect("notify::album", self.update_subtitle),
        ]
        self.update_title()
        self.update_subtitle()

    def unbind(self, *args):
        for conn in self._connections:
            self.recording.disconnect(conn)

        self.recording = None
        self.parent = None

    def update_title(self, *args):
        title = html.escape(self.recording.title)
        if self.recording.disambiguation:
            opacity = 0.55
            if Adw.StyleManager.get_default().get_high_contrast():
                opacity = 0.9
            title += (
                f' <span alpha="{int(opacity * 100)}%">('
                + html.escape(self.recording.disambiguation)
                + ")</span>"
            )
        self.set_title(title)

    def update_subtitle(self, *args):
        self._subtitle = f"{self.recording.artist or 'N/A'} • {self.recording.album or 'N/A'} ({self.file_name or 'N/A'})"  # noqa: E501
        self.set_subtitle(html.escape(self._subtitle))

    def start_loading(self):
        self.suffix_stack.set_visible(True)
        self.suffix_stack.set_visible_child(self.loading_icon)

    def mark_as_unidentified(self):
        self.suffix_stack.set_visible_child(self.not_found_icon)

    @Gtk.Template.Callback()
    def toggle_apply(self, toggle, *args):
        if toggle.props.active and self.file_id not in self.parent.parent.apply_files:
            self.parent.parent.apply_files.append(self.file_id)
            self.parent.parent.apply_files_changed()
        elif not toggle.props.active and self.file_id in self.parent.parent.apply_files:
            self.parent.parent.apply_files.remove(self.file_id)
            self.parent.parent.apply_files_changed()

    @Gtk.Template.Callback()
    def open_recording_url(self, *args):
        """Open the currently bound recording's MusicBrainz page."""

        async def _open(self):
            self.recording_url_launcher.props.uri = (
                f"https://musicbrainz.org/recording/{self.recording.recording_id}"
            )
            await self.recording_url_launcher.launch(self.get_root())

        asyncio.create_task(_open(self))

    def get_mbid_from_recording_override_entry(self) -> str | None:
        """
        Extract the MBID from the URL or raw ID string given to the recording
        override entry.
        """
        entry_text = self.recording_override_entry.get_text().strip()

        if re.match(f"^{MBID_REGEX}$", entry_text):
            # Raw MBID
            mbid = entry_text
        else:
            # URL
            match = re.match(RECORDING_URL_REGEX, entry_text)
            try:
                mbid = match.group("mbid")
            except (AttributeError, IndexError):  # no match
                return None

        return mbid

    @Gtk.Template.Callback()
    def validate_recording_override_entry(self, *args):
        """Validate the contents of the recording override entry."""
        if self.get_mbid_from_recording_override_entry():
            self.recording_override_entry.remove_css_class("error")
            self.recording_override_apply_button.props.sensitive = True
        else:
            self.recording_override_entry.add_css_class("error")
            self.recording_override_apply_button.props.sensitive = False

    @Gtk.Template.Callback()
    def set_recording_override_from_entry(self, *args):
        """Apply the recording ID or URL from the recording override entry."""

        async def _set_rec_override(self):
            mbid = self.get_mbid_from_recording_override_entry()
            if not mbid:
                return

            self.recording_override_popover.props.sensitive = False

            rec = await MusicBrainzRecording.new_for_id(mbid)

            self.parent.parent._identify_set_recording(self.file_id, rec)
            self.parent.parent.on_identify_done(None, show_toast=False)

            self.recording_override_popover.props.sensitive = True

        asyncio.create_task(_set_rec_override(self))


@Gtk.Template(resource_path=f"{APP_GRESOURCE_PATH}/ui/identify/identify.ui")
class EartagIdentifyDialog(Adw.Dialog):
    __gtype_name__ = "EartagIdentifyDialog"

    toast_overlay = Gtk.Template.Child()

    id_progress = Gtk.Template.Child()
    content_listbox = Gtk.Template.Child()

    cancel_button = Gtk.Template.Child()
    identify_button = Gtk.Template.Child()
    apply_button = Gtk.Template.Child()
    end_button_stack = Gtk.Template.Child()

    apply_progress = GObject.Property(type=float, minimum=0, maximum=1)

    def __init__(self, window):
        super().__init__()
        self.parent = window
        self.file_manager = window.file_manager

        self.files = Gio.ListStore(item_type=EartagFile)
        self.unidentified_filter = Gtk.CustomFilter()
        self.unidentified_filter.set_filter_func(self.unidentified_filter_func)
        self.unidentified_filter.connect("changed", self.mark_unidentified_filter_as_changed)
        self.files_unidentified = Gtk.FilterListModel(
            model=self.files, filter=self.unidentified_filter
        )
        self.recordings = {}  # file.id: MusicBrainzRecording
        self.recordings_model = Gio.ListStore(item_type=MusicBrainzRecording)
        self.apply_files = []
        self.release_rows = {}  # release.id: EartagIdentifyReleaseRow

        self.identify_task = EartagAsyncTask(self.identify_files)
        self.apply_task = EartagAsyncTask(self.apply_func)

        # For some reason we can't create this from the template, so it
        # has to be added here:
        self.unidentified_row = EartagModelExpanderRow()
        self.unidentified_row.set_title(_("Unidentified Files"))
        self.unidentified_row.set_expanded(True)
        cover_dummy = EartagIdentifyCoverImage()
        cover_dummy.set_hexpand(False)
        self.unidentified_row.add_prefix(cover_dummy)
        self.unidentified_row.bind_model(self.files_unidentified, self.unidentified_row_create)

        self.content_listbox.append(self.unidentified_row)

        self.identify_task.bind_property("progress", self.id_progress, "fraction")
        self.identify_task.connect("task-done", self.on_identify_done)

        self.apply_task.bind_property("progress", self.id_progress, "fraction")
        self.apply_task.connect("task-done", self.on_apply_done)

        self.files.splice(0, self.files.get_n_items(), self.file_manager.selected_files_list)

        self.connect("closed", self.on_close_request)

    def on_close_request(self, *args):
        for row in self.release_rows.values():
            row.unbind()
        self.release_rows = {}

    def unidentified_row_create(self, file, *args):
        return EartagIdentifyFileRow(file, self)

    @Gtk.Template.Callback()
    def on_cancel(self, *args):
        try:
            if self.identify_task.is_running:
                self.identify_task.stop()
            if self.apply_task.is_running:
                self.apply_task.stop()
        except AttributeError:
            pass
        self.files = None
        self.props.can_close = True
        self.close()

    @Gtk.Template.Callback()
    def do_identify(self, *args):
        self.identify_button.set_sensitive(False)
        self.apply_button.set_sensitive(False)
        self.set_can_close(False)

        self.identify_task.run()

    def _identify_set_recording(self, file_id: str, rec: MusicBrainzRecording):
        try:
            rec.release  # noqa: B018
        except ValueError:
            rec.release = rec.available_releases[0]

        if file_id in self.recordings:
            old_rec = self.recordings[file_id]
            index = self.recordings_model.find(old_rec)[1]
            if index is not None:
                self.recordings_model.remove(index)
            old_relrow = self.release_rows[old_rec.release.release_id]
            old_relrow._filter_changed = False
            old_relrow.update_filter()
            old_relrow._filter_changed = False
            if old_relrow._rec_sorter_model.get_n_items() == 0:
                self.content_listbox.remove(old_relrow)
                del self.release_rows[old_rec.release.release_id]

        self.recordings[file_id] = rec
        self.recordings_model.append(rec)

        if rec.release.release_id in self.release_rows:
            row = self.release_rows[rec.release.release_id]
            row._filter_changed = False
            row.update_filter()
            row._filter_changed = False
        else:
            self.release_rows[rec.release.release_id] = EartagIdentifyReleaseRow(self, rec.release)
            self.content_listbox.prepend(self.release_rows[rec.release.release_id])

        self._filter_changed = False
        self.unidentified_filter.changed(Gtk.FilterChange.DIFFERENT)
        self._filter_changed = False

        self.apply_files.append(file_id)

    async def identify_files(self, *args, **kwargs):
        progress_step = 1 / len(self.files)

        found_recordings = []

        # Step 1. Go over each file and find recordings
        for file in self.files:
            logger.info(f"Identifying recordings for file: {file}")
            unid_index = find_in_model(self.files_unidentified, file)
            if unid_index < 0:
                logger.error("Could not find file in unidentifed filter, this should never happen!")
                continue

            unid_row = self.unidentified_row.get_row_at_index(unid_index)
            unid_row.start_loading()

            recordings = []

            try:
                if file.title and file.artist:
                    # For files with a title/artist tag, look up the data in MusicBrainz
                    recordings = await MusicBrainzRecording.get_recordings_for_file(file)
                else:
                    logger.debug("No title or artist tag, trying to guess tags from filename...")
                    tag_guesses = []
                    for guess_pattern in (
                        "{artist} - {title}",
                        "{title} - {artist}",
                        "{artist} — {title}",
                        "{title} — {artist}",
                        "{title}",
                    ):
                        tag_guesses.append(
                            extract_tags_from_filename(
                                file.path, guess_pattern, strip_common_suffixes=True
                            )
                        )

                    for guess in tag_guesses:
                        if not guess:
                            continue

                        recordings += await MusicBrainzRecording.get_recordings_for_file(
                            file, overrides=guess, sort=False
                        )

                    recordings = MusicBrainzRecording.sort_recordings(recordings, file=file)

                # If we don't find anything or we find multiple recordings, try AcoustID
                if (
                    not recordings
                    or len(recordings) > 1
                    or (recordings[0].release and recordings[0].release.status != "official")
                    or not file.title
                    or not file.artist
                ):
                    logger.debug("Trying AcoustID, just to be sure...")
                    id_confidence, id_recording = await acoustid_identify_file(file)

                    if id_recording:
                        recordings = MusicBrainzRecording.sort_recordings(
                            recordings + [id_recording], file=file
                        )

                if recordings:
                    rec = recordings[0]
                    found_recordings.append(rec)
                    if not rec.available_releases:
                        unid_row.mark_as_unidentified()
                    else:
                        self._identify_set_recording(file.id, rec)
                        self.identify_task.increment_progress(progress_step)
                else:
                    unid_row.mark_as_unidentified()

            except asyncio.exceptions.CancelledError:
                return

            except:  # noqa: E722
                # Prevent crashes in recording identification from aborting the identification
                logger.error(f"Failure while identifying file {file}:")
                traceback.print_exc()
                unid_row.mark_as_unidentified()
                continue

        # Once we're done identifying all recordings, go back through found releases and
        # group together releases belonging to the same release group.
        #
        # We need to do this since we might be dealing with a "deluxe edition" with extra
        # tracks; in that case, all tracks should fall under the deluxe edition.

        logger.debug("Done identifying files, grouping will now proceed")

        groups = {}  # group ID: releases for this group
        for rel in [row.release for row in self.release_rows.values()]:
            if rel.group.relgroup_id not in groups:
                groups[rel.group.relgroup_id] = [rel]
            else:
                groups[rel.group.relgroup_id].append(rel)

        logger.debug(f"Per-group identified releases: {groups}")

        group_recs = {}  # group ID: recordings for this group
        for rec in found_recordings:
            if rec.release.group.relgroup_id in group_recs:
                group_recs[rec.release.group.relgroup_id].append(rec)
            else:
                group_recs[rec.release.group.relgroup_id] = [rec]

        logger.debug(f"Per-group identified recordings: {group_recs}")

        # For each group, figure out which one contains the most of our identified
        # recordings. The one that covers the most wins.
        for group_id in groups.keys():
            rel_trackcount = {}  # release ID: [recording IDs]
            for rec in group_recs[group_id]:
                rec_rel_ids = [rel.release_id for rel in rec.available_releases]
                for rel_id in rec_rel_ids:
                    if rel_id in rel_trackcount:
                        rel_trackcount[rel_id].add(rec.recording_id)
                    else:
                        rel_trackcount[rel_id] = {
                            rec.recording_id,
                        }

            rel_trackcount_sorted = sorted(
                rel_trackcount.items(), key=lambda v: len(v[1]), reverse=True
            )

            logger.debug(f"Found {len(rel_trackcount_sorted)} possible releases")

            target_release = rel_trackcount_sorted[0][0]
            target_rel_recordings = set()

            logger.debug(f"Selected {target_release} as target release")

            for rec in group_recs[group_id]:
                for rel in rec.available_releases:
                    if rel.release_id == target_release:
                        target_rel_recordings.add(rec.recording_id)
                        rec.release = rel
                        break

            # Set up alternative releases
            alt_releases = set()
            for rel_id, rec_ids in rel_trackcount.items():
                if rel_trackcount[target_release].issubset(rec_ids):
                    alt_releases.add(rel_id)

            logger.debug(
                f"Found {len(alt_releases)} alternative releases for release {target_release}:\n{alt_releases}"
            )
            all_releases = set()
            for rec in group_recs[group_id]:
                if rec.recording_id not in target_rel_recordings:
                    continue
                all_releases = all_releases.union(set(rec.available_releases))

            self.release_rows[target_release].refresh_alternative_releases(
                [rel for rel in all_releases if rel.release_id in alt_releases]
            )

        for k, row in list(self.release_rows.items()).copy():
            row.update_filter()
            if row._rec_filter_model.get_n_items() == 0:
                del self.release_rows[k]
                self.content_listbox.remove(row)

    def on_identify_done(self, task, *args, show_toast: bool = True):
        try:
            identified = self.files.get_n_items() - self.files_unidentified.get_n_items()
        except AttributeError:  # this happens when the operation is cancelled
            return
        self.end_button_stack.set_visible_child(self.apply_button)
        self.apply_button.set_sensitive(bool(identified))
        self.set_can_close(True)
        for relrow in self.release_rows.values():
            relrow.toggle_apply_sensitivity(True)
        if not self.files_unidentified.get_n_items():
            self.unidentified_row.set_visible(False)

        if show_toast:
            self.toast_overlay.add_toast(
                Adw.Toast.new(
                    ngettext(
                        # TRANSLATORS: {identified} is a placeholder for the number
                        # of tracks that were succesfully identified.
                        # **Do not translate the text between the curly brackets!**
                        "Identified 1 track",
                        "Identified {identified} tracks",
                        identified,
                    ).format(identified=identified)
                )
            )

    @Gtk.Template.Callback()
    def do_apply(self, *args):
        self.apply_button.set_sensitive(False)
        for relrow in self.release_rows.values():
            relrow.toggle_apply_sensitivity(False)
        self.set_can_close(False)

        self.apply_task.run()

    async def apply_func(self):
        files = [file for file in self.files if file.id in self.apply_files]
        if not files:
            return

        # Step 1. Download covers
        releases = set()
        for file in files:
            rec = self.recordings[file.id]
            releases.add(rec.release)

        progress_step = 1 / (len(files) + len(releases))

        sem = asyncio.Semaphore(5)
        _tasks = set()
        async with asyncio.TaskGroup() as tg:
            while releases:
                async with sem:
                    _tasks.add(tg.create_task(releases.pop().download_covers_async()))
                    self.apply_task.increment_progress(progress_step)
        del releases
        del _tasks
        del sem

        # Step 2. Apply tags to files
        for file in files:
            rec = self.recordings[file.id]
            # Ugly quirk: since release objects are derived from the data in the
            # recording object, and not shared, downloading covers for one does
            # not mark the covers for all recordings as downloaded.
            #
            # We run download_covers_async() on each of the recordings to update
            # the cover data; no actual data gets downloaded, since covers are
            # cached.
            await rec.download_covers_async()

            await rec.apply_data_to_file(file)
            self.apply_task.increment_progress(progress_step)

    def on_apply_done(self, *args):
        self.props.can_close = True
        self.file_manager.emit("refresh-needed")
        try:
            identified = self.files.get_n_items() - self.files_unidentified.get_n_items()
        except AttributeError:  # this happens when the operation is cancelled
            return
        self.parent.toast_overlay.add_toast(
            Adw.Toast.new(
                ngettext(
                    # TRANSLATORS: {n} is a placeholder for the number
                    # of tracks that were succesfully identified.
                    # **Do not translate the text between the curly brackets!**
                    "Applied changes to 1 track",
                    "Applied changes to {n} tracks",
                    identified,
                ).format(n=identified)
            )
        )
        self.files = None
        self.identify_task = None
        self.apply_task = None
        self.close()

    def unidentified_filter_func(self, file, *args):
        return file.id not in self.recordings

    def apply_files_changed(self, *args):
        self.apply_button.set_sensitive(bool(self.apply_files))

    def mark_unidentified_filter_as_changed(self, *args):
        """
        Since we call a filter change from a thread, we have to wait for it
        to finish first to allow the unidentified rows to refresh, else we
        get weird errors in the identify loop.
        """
        self._filter_changed = True
